#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import io
import os
import re
import dataclasses
import pathlib
import tarfile
import traceback
from typing import Union, Optional, List, Tuple

import shutil

import numpy as np

from ase import Atoms
from ase.io import read, write
from ase.calculators.vasp import Vasp

from ..data.extatoms import ScfErrAtoms
from ..utils.strucopy import read_sort, resort_atoms_with_spc
from ..utils.cmdrun import run_ase_calculator
from .driver import AbstractDriver, DriverSetting, Controller


"""Driver for VASP."""
#: Ase-vasp resort fname.
ASE_VASP_SORT_FNAME: str = "ase-sort.dat"


def read_outcar_scf(lines: List[str]) -> Tuple[int, float]:
    """"""
    nelm, ediff = None, None
    for line in lines:
        if line.strip().startswith("NELM"):
            nelm = int(line.split()[2][:-1])
        if line.strip().startswith("EDIFF"):
            ediff = float(line.split()[2])
        if nelm is not None and ediff is not None:
            break
    else:
        ...  # TODO: raise an error?

    return nelm, ediff


def read_oszicar(lines: List[str], nelm: int, ediff: float) -> List[bool]:
    """"""
    convergence = []
    content = ""
    for line in lines:
        start = line.strip().split()[0]
        if start == "N":
            content = ""
            continue
        if start.isdigit():
            scfsteps = [int(s.split()[1]) for s in content.strip().split("\n")]
            num_scfsteps = len(scfsteps)
            assert num_scfsteps == scfsteps[-1]
            enediffs = [float(s.split()[3]) for s in content.strip().split("\n")]
            is_converged = num_scfsteps < nelm or np.fabs(enediffs[-1]) <= ediff
            convergence.append(is_converged)
        content += line

    # If the last SCF is not finished,
    # there is no content to check

    return convergence


@dataclasses.dataclass
class LangevinThermostat(Controller):

    name: str = "langevin"

    def __post_init__(
        self,
    ):
        """"""
        friction = self.params.get("friction", None)  # fs^-1
        assert friction is not None

        self.conv_params = dict(langevin_gamma=friction * 1e3)  # ps^-1

        return


@dataclasses.dataclass
class NoseHooverThermostat(Controller):

    name: str = "nose_hoover"

    def __post_init__(
        self,
    ):
        """"""
        # FIXME: convert taut to smass
        smass = self.params.get("taut", 0.0)  # or smass?
        assert smass >= 0, "NoseHoover-NVT needs positive SMASS."

        self.conv_params = dict(smass=smass)

        return


@dataclasses.dataclass
class ParrinelloRahmanBarostat(Controller):

    name: str = "parrinello_rahman"

    def __post_init__(
        self,
    ):
        """"""
        # FIXME: convert taut to smass
        smass = self.params.get("taut", 0.0)  # or smass?
        assert smass >= 0, f"{self.name} needs positive SMASS."

        friction = self.params.get("friction", None)  # fs^-1
        assert friction is not None

        friction_lattice = self.params.get("friction_lattice", None)  # fs^-1
        assert friction_lattice is not None

        pmass = self.params.get("pmass", None)  # a.m.u
        assert pmass > 0.0

        self.conv_params = dict(
            smass=smass,
            langevin_gamma=friction * 1e3,  # array, ps^-1
            langevin_gamma_l=friction_lattice * 1e3,  # real, ps^-1
            pmass=1000.0,
        )

        return


controllers = dict(
    langevin_nvt=LangevinThermostat,
    nose_hoover_nvt=NoseHooverThermostat,
    parrinello_rahman_npt=ParrinelloRahmanBarostat,
)


@dataclasses.dataclass
class VaspDriverSetting(DriverSetting):

    # - md setting
    ensemble: str = "nve"

    # - driver detailed controller setting
    controller: dict = dataclasses.field(default_factory=dict)

    # TODO: move below to controller?
    # fix_cm: bool = False

    # - min setting
    etol: float = None
    fmax: float = 0.05

    def __post_init__(self):
        """Convert parameters into driver-specific ones.

        These parameters are frozen when the driver is initialised.

        """
        # - update internals that are specific for each calculator...
        if self.task == "min":
            # minimisation
            if self.min_style == "bfgs":
                ibrion = 1
            elif self.min_style == "cg":
                ibrion = 2
            else:
                # raise ValueError(f"Unknown minimisation {self.min_style} for vasp".)
                ...

            self._internals.update(ibrion=ibrion, potim=self.maxstep)

        # -- cmin: cell minimisation
        if self.task == "cmin":
            if self.min_style == "bfgs":
                ibrion = 1
            elif self.min_style == "cg":
                ibrion = 2
            else:
                # raise ValueError(f"Unknown minimisation {self.min_style} for vasp".)
                ...

            self._internals.update(isif=3, ibrion=ibrion, potim=self.maxstep)

        if self.task == "md":
            # NOTE: Always use Selective Dynamics and MDALAGO
            #       since it properly treats the DOF and velocities
            # some general
            # if self.velocity_seed is None:
            #     self.velocity_seed = np.random.randint(0, 10000)
            # random_seed = [self.velocity_seed, 0, 0]
            self._internals.update(
                # -- Some shared parameters, every MD needs these!!
                velocity_seed=self.velocity_seed,
                ignore_atoms_velocities=self.ignore_atoms_velocities,
                ibrion=0,
                isif=0,
                potim=self.timestep,  # fs
                random_seed=None,  # NOTE: init later in driver run
            )

            if self.ensemble == "nve":
                _init_md_params = dict()
                _init_md_params.update(
                    mdalgo=2,
                    smass=-3,
                )
            elif self.ensemble == "nvt":
                # We need keywords: TEBEG and TEEND.
                # Also, only consistent-temperature simulation is supported.
                _init_md_params = dict(tebeg=self.temp, teend=self.temp)

                if self.controller is not None:
                    thermo_cls_name = self.controller["name"] + "_" + self.ensemble
                    thermo_cls = controllers[thermo_cls_name]
                else:
                    thermo_cls = LangevinThermostat
                thermostat = thermo_cls(**self.controller)

                if thermostat.name == "langevin":
                    # MDALGO, LANGEVIN_GAMMA
                    _init_md_params.update(
                        mdalgo=3,
                    )
                    _init_md_params.update(**thermostat.conv_params)
                elif thermostat.name == "nose_hoover":
                    # MDALGO, SMASS
                    _init_md_params.update(
                        mdalgo=2,
                    )
                    _init_md_params.update(**thermostat.conv_params)
                else:
                    raise RuntimeError(f"Unknown {thermostat =}.")
            elif self.ensemble == "npt":
                # We need keywords: TEBEG, TEEND, PSTRESS
                _init_md_params = dict(
                    tebeg=self.temp,
                    teend=self.temp,
                    # pressure unit 1 GPa  = 10 kBar
                    #               1 kBar = 1000 bar = 10^8 Pa
                    pstress=1e-3 * self.press,  # vasp uses kB
                )

                if self.controller is not None:
                    baro_cls_name = self.controller["name"] + "_" + self.ensemble
                    baro_cls = controllers[baro_cls_name]
                else:
                    baro_cls = ParrinelloRahmanBarostat
                barostat = baro_cls(**self.controller)

                if barostat.name == "parrinello_rahman":
                    _init_md_params.update(mdalgo=3)
                    _init_md_params.update(**barostat.conv_params)
                else:
                    raise RuntimeError(f"Unknown {barostat =}.")
            else:
                raise NotImplementedError(f"{self.md_style} is not supported yet.")

            self._internals.update(**_init_md_params)

        if self.task == "freq":
            # ibrion, nfree, potim
            raise NotImplementedError("")

        return

    def get_run_params(self, *args, **kwargs):
        """"""
        # convergence criteria
        fmax_ = kwargs.get("fmax", self.fmax)
        etol_ = kwargs.get("etol", self.etol)

        # etol is prioritised
        if etol_ is not None:
            ediffg = etol_
        else:
            if fmax_ is not None:
                ediffg = -1.0 * fmax_
            else:
                ediffg = -5e-2

        steps_ = kwargs.get("steps", self.steps)
        nsw = steps_

        run_params = dict(
            constraint=kwargs.get("constraint", self.constraint), ediffg=ediffg, nsw=nsw
        )

        return run_params


class VaspDriver(AbstractDriver):

    name = "vasp"

    # - defaults
    default_task = "min"
    supported_tasks = ["min", "cmin", "md", "freq"]

    # - system depandant params
    syswise_keys: List[str] = ["system", "kpts", "kspacing"]

    # - file names would be copied when continuing a calculation
    saved_fnames = [
        "ase-sort.dat",
        "INCAR",
        "POSCAR",
        "KPOINTS",
        "POTCAR",
        "OSZICAR",
        "OUTCAR",
        "CONTCAR",
        "vasprun.xml",
        "REPORT",
    ]

    def __init__(self, calc: Vasp, params: dict, directory="./", *args, **kwargs):
        """"""
        super().__init__(calc, params, directory=directory, *args, **kwargs)

        self.setting = VaspDriverSetting(**params)

        return

    def _verify_checkpoint(self, *args, **kwargs) -> bool:
        """Check whether there is a previous calculation in the `self.directory`."""
        verified = True
        if self.directory.exists():
            vasprun = self.directory / "vasprun.xml"
            if vasprun.exists() and vasprun.stat().st_size != 0:
                temp_frames = read(vasprun, ":")
                try:
                    _ = temp_frames[0].get_forces()
                except:  # `RuntimeError: Atoms object has no calculator.`
                    verified = False
            else:
                verified = False
        else:
            verified = False

        return verified

    def _irun(
        self,
        atoms: Atoms,
        ckpt_wdir=None,
        cache_traj: Optional[List[Atoms]] = None,
        *args,
        **kwargs,
    ):
        """"""
        try:
            if ckpt_wdir is None:  # start from the scratch
                # - merge params
                run_params = self.setting.get_run_params(**kwargs)
                run_params.update(**self.setting.get_init_params())
                run_params["system"] = self.directory.name

                # FIXME: Init velocities?
                prev_ignore_atoms_velocities = run_params.pop(
                    "ignore_atoms_velocities", False
                )
                velocity_seed = run_params.pop("velocity_seed", None)

                if self.setting.task == "md":
                    vasp_random_seed = [self.random_seed, 0, 0]
                    self._print(f"MD Driver's velocity_seed: vasp-{vasp_random_seed}")
                    self._print(f"MD Driver's rng: vasp-{vasp_random_seed}")
                    run_params["random_seed"] = vasp_random_seed
                    # TODO: use external velocities?
                else:
                    ...

                # - update some system-dependant params
                if "langevin_gamma" in run_params:
                    ntypes = len(set(atoms.get_chemical_symbols()))
                    run_params["langevin_gamma"] = [
                        run_params["langevin_gamma"]
                    ] * ntypes

                # FIXME: LDA+U

                # - constraint
                self._preprocess_constraints(atoms, run_params)

                self.calc.set(**run_params)
                atoms.calc = self.calc
                # NOTE: ASE VASP does not write velocities and thermostat to POSCAR
                #       thus we manually call the function to write input files and
                #       run the calculation
                self.calc.write_input(atoms)
            else:
                self.calc.read_incar(ckpt_wdir / "INCAR")  # read previous incar
                if cache_traj is None:
                    traj = self.read_trajectory()
                else:
                    traj = cache_traj
                nframes = len(traj)
                assert nframes > 0, "VaspDriver restarts with a zero-frame trajectory."
                dump_period = 1  # since we read vasprun.xml, every frame is dumped
                target_steps = self.setting.get_run_params(*args, **kwargs)["nsw"]
                if target_steps > 0:  # not a spc
                    # BUG: ...
                    steps = target_steps + dump_period - nframes * dump_period
                    assert (
                        steps > 0
                    ), f"Steps should be greater than 0. (steps = {steps})"
                    self.calc.set(nsw=steps)
                # NOTE: ASE VASP does not write velocities and thermostat to POSCAR
                #       thus we manually call the function to write input files and
                #       run the calculation
                # FIXME: Read random_seed in REPORT!!!
                self.calc.write_input(atoms)
                # To restart, velocities are always retained
                # if (self.directory/"CONTCAR").exists() and (self.directory/"CONTCAR").stat().st_size != 0:
                #    shutil.copy(self.directory/"CONTCAR", self.directory/"POSCAR")
                shutil.copy(ckpt_wdir / "CONTCAR", self.directory / "POSCAR")

            run_ase_calculator("vasp", self.calc.command, self.directory)

        except Exception as e:
            self._debug(f"Exception of {self.__class__.__name__} is {e}.")
            self._debug(
                f"Exception of {self.__class__.__name__} is {traceback.format_exc()}."
            )

        return

    def _read_a_single_trajectory(self, wdir, archive_path=None, *args, **kwargs):
        """"""
        oszicar_lines, outcar_lines = None, None
        if archive_path is None:
            # - read trajectory
            vasprun = wdir / "vasprun.xml"
            frames = read(vasprun, ":")
            with open(wdir / "OUTCAR") as fopen:
                outcar_lines = fopen.readlines()
            with open(wdir / "OSZICAR") as fopen:
                oszicar_lines = fopen.readlines()
        else:
            flags = [False, False, False]
            vasprun_name = str(
                (wdir / "vasprun.xml").relative_to(self.directory.parent)
            )
            oszicar_name = str((wdir / "OSZICAR").relative_to(self.directory.parent))
            outcar_name = str((wdir / "OUTCAR").relative_to(self.directory.parent))
            with tarfile.open(archive_path, "r:gz") as tar:
                for tarinfo in tar:
                    if tarinfo.name == vasprun_name:
                        fobj = io.StringIO(
                            tar.extractfile(tarinfo.name).read().decode()
                        )
                        frames = read(fobj, ":", format="vasp-xml")
                        fobj.close()
                        flags[0] = True
                    if tarinfo.name == oszicar_name:
                        fobj = io.StringIO(
                            tar.extractfile(tarinfo.name).read().decode()
                        )
                        oszicar_lines = fobj.readlines()
                        fobj.close()
                        flags[1] = True
                    if tarinfo.name == outcar_name:
                        fobj = io.StringIO(
                            tar.extractfile(tarinfo.name).read().decode()
                        )
                        outcar_lines = fobj.readlines()
                        fobj.close()
                        flags[2] = True
                    if all(flags):
                        break
                else:  # TODO: if not find target traj?
                    ...

        # - read oszicar and outcar
        if outcar_lines is not None and oszicar_lines is not None:
            nelm, ediff = read_outcar_scf(outcar_lines)
            scf_convergences = read_oszicar(oszicar_lines, nelm, ediff)
            assert len(scf_convergences) == len(frames)
            for i, is_converged in enumerate(scf_convergences):
                if not is_converged:
                    frames[i] = ScfErrAtoms.from_atoms(frames[i])
                    self._print(f"ScfErrAtoms Step {i} @ {str(wdir)}")

        return frames

    def read_trajectory(
        self, add_step_info=True, archive_path=None, *args, **kwargs
    ) -> List[Atoms]:
        """Read trajectory in the current working directory.

        If the calculation failed, an empty atoms with errof info would be returned.

        """
        # - read structures
        # -- read backups
        prev_wdirs = []
        if archive_path is None:
            prev_wdirs = sorted(self.directory.glob(r"[0-9][0-9][0-9][0-9][.]run"))
        else:
            pattern = self.directory.name + "/" + r"[0-9][0-9][0-9][0-9][.]run"
            with tarfile.open(archive_path, "r:gz") as tar:
                for tarinfo in tar:
                    if tarinfo.isdir() and re.match(pattern, tarinfo.name):
                        prev_wdirs.append(tarinfo.name)
            prev_wdirs = [
                self.directory / pathlib.Path(p).name for p in sorted(prev_wdirs)
            ]
        self._debug(f"prev_wdirs: {prev_wdirs}")

        traj_list = []
        for w in prev_wdirs:
            curr_frames = self._read_a_single_trajectory(w, archive_path=archive_path)
            traj_list.append(curr_frames)

        # Even though vasprun file may be empty, the read can give a empty list...
        vasprun = self.directory / "vasprun.xml"
        curr_frames = self._read_a_single_trajectory(
            self.directory, archive_path=archive_path
        )
        traj_list.append(curr_frames)

        # -- concatenate
        # NOTE: Some spin systems may give different scf convergence on the same
        #       structure. Sometimes, the preivous failed but the next run converged,
        #       The concat below uses the previous one...
        traj_frames_, ntrajs = [], len(traj_list)
        if ntrajs > 0:
            traj_frames_.extend(traj_list[0])
            for i in range(1, ntrajs):
                assert np.allclose(
                    traj_list[i - 1][-1].positions, traj_list[i][0].positions
                ), f"Traj {i-1} and traj {i} are not consecutive."
                traj_frames_.extend(traj_list[i][1:])
        else:
            ...

        nframes = len(traj_frames_)
        natoms = len(traj_frames_[0])

        # - sort frames
        traj_frames = []
        if nframes > 0:
            if (self.directory / ASE_VASP_SORT_FNAME).exists():
                sort, resort = read_sort(self.directory)
            else:  # without sort file, use default order
                sort, resort = list(range(natoms)), list(range(natoms))
            for i, sorted_atoms in enumerate(traj_frames_):
                # NOTE: calculation with only one unfinished step does not have forces
                input_atoms = resort_atoms_with_spc(
                    sorted_atoms,
                    resort,
                    "vasp",
                    print_func=self._print,
                    debug_func=self._debug,
                )
                # if input_atoms is None:
                #    input_atoms = Atoms()
                #    input_atoms.info["error"] = str(self.directory)
                if input_atoms is not None:
                    if add_step_info:
                        input_atoms.info["step"] = i
                    traj_frames.append(input_atoms)
        else:
            ...

        return traj_frames


if __name__ == "__main__":
    ...



"""ASE calculator interface module."""

import copy
from pathlib import (
    Path,
)
from typing import (
    TYPE_CHECKING,
    Dict,
    List,
    Optional,
    Union,
)

import numpy as np

from ase.calculators.calculator import (
    Calculator,
    PropertyNotImplementedError,
    all_changes,
)

from deepmd import (
    DeepPotential,
)

if TYPE_CHECKING:
    from ase import (
        Atoms,
    )

__all__ = ["DP"]


class DP(Calculator):
    """Implementation of ASE deepmd calculator.

    Implemented propertie are `energy`, `forces` and `stress`

    Parameters
    ----------
    model : Union[str, Path]
        path to the model
    label : str, optional
        calculator label, by default "DP"
    type_dict : Dict[str, int], optional
        mapping of element types and their numbers, best left None and the calculator
        will infer this information from model, by default None

    Examples
    --------
    Compute potential energy

    >>> from ase import Atoms
    >>> from deepmd.calculator import DP
    >>> water = Atoms('H2O',
    >>>             positions=[(0.7601, 1.9270, 1),
    >>>                        (1.9575, 1, 1),
    >>>                        (1., 1., 1.)],
    >>>             cell=[100, 100, 100],
    >>>             calculator=DP(model="frozen_model.pb"))
    >>> print(water.get_potential_energy())
    >>> print(water.get_forces())

    Run BFGS structure optimization

    >>> from ase.optimize import BFGS
    >>> dyn = BFGS(water)
    >>> dyn.run(fmax=1e-6)
    >>> print(water.get_positions())
    """

    name = "DP"
    implemented_properties = ["energy", "free_energy", "forces", "virial", "stress"]

    def __init__(
        self,
        model: Union[str, "Path"],
        label: str = "DP",
        type_dict: Dict[str, int] = None,
        **kwargs,
    ) -> None:
        Calculator.__init__(self, label=label, **kwargs)

        self.model_path = str(Path(model).resolve())
        self.type_dict = type_dict

        # - lazy init
        self.dp = None # will init when first calculate

    def calculate(
        self,
        atoms: Optional["Atoms"] = None,
        properties: List[str] = ["energy", "forces", "virial"],
        system_changes: List[str] = all_changes,
    ):
        """Run calculation with deepmd model.

        Parameters
        ----------
        atoms : Optional[Atoms], optional
            atoms object to run the calculation on, by default None
        properties : List[str], optional
            unused, only for function signature compatibility,
            by default ["energy", "forces", "stress"]
        system_changes : List[str], optional
            unused, only for function signature compatibility, by default all_changes
        """
        if atoms is not None:
            self.atoms = atoms.copy()

        if self.dp is None:
            self.dp = DeepPotential(self.model_path)
            if self.type_dict is None:
                self.type_dict = dict(
                    zip(self.dp.get_type_map(), range(self.dp.get_ntypes()))
                )

        coord = self.atoms.get_positions().reshape([1, -1])
        if sum(self.atoms.get_pbc()) > 0:
            cell = self.atoms.get_cell().reshape([1, -1])
        else:
            cell = None
        symbols = self.atoms.get_chemical_symbols()
        atype = [self.type_dict[k] for k in symbols]
        e, f, v = self.dp.eval(coords=coord, cells=cell, atom_types=atype)
        self.results["energy"] = e[0][0]
        # see https://gitlab.com/ase/ase/-/merge_requests/2485
        self.results["free_energy"] = e[0][0]
        self.results["forces"] = f[0]
        self.results["virial"] = v[0].reshape(3, 3)

        # convert virial into stress for lattice relaxation
        if "stress" in properties:
            if sum(atoms.get_pbc()) > 0:
                # the usual convention (tensile stress is positive)
                # stress = -virial / volume
                stress = -0.5 * (v[0].copy() + v[0].copy().T) / atoms.get_volume()
                # Voigt notation
                self.results["stress"] = stress.flat[[0, 4, 8, 5, 2, 1]]
            else:
                raise PropertyNotImplementedError


class BatchDP:

    name = "DP"
    implemented_properties = ["energy", "free_energy", "forces", "virial", "stress"]

    """Evaluate a batch of structures.

    This is useful in ASE NEB calculation.

    """

    def __init__(
        self,
        model: Union[str, "Path"],
        label: str = "DP",
        type_dict: Dict[str, int] = None,
        **kwargs,
    ) -> None:
        """"""
        self.model_path = str(Path(model).resolve())
        self.type_dict = type_dict

        # - lazy init
        self.dp = None # will init when first calculate

        return
    
    def calculate(
        self,
        frames: Optional[List["Atoms"]] = None,
        properties: List[str] = ["energy", "forces", "virial"],
        system_changes: List[str] = all_changes,
    ):
        """"""
        if self.dp is None:
            self.dp = DeepPotential(self.model_path)
            if self.type_dict is None:
                self.type_dict = dict(
                    zip(self.dp.get_type_map(), range(self.dp.get_ntypes()))
                )
        
        # - convert atoms
        # TODO: assume it is not mixed-type
        coords, cells, atypes = [], [], []
        for atoms in frames:
            curr_coord, curr_cell, curr_atype = self._convert_from_ase_to_dp(atoms)
            coords.append(curr_coord)
            cells.append(curr_cell)
            atypes.append(curr_atype)
        coords = np.array(coords)
        cells = np.array(cells)
        atype = atypes[0]

        # - calc
        e, f, v = self.dp.eval(coords=coords, cells=cells, atom_types=atype)

        # - convert results
        self.results = {}
        self.results["energy"] = [x[0] for x in e] # List[float]
        self.results["free_energy"] = [x[0] for x in e] # List[float]
        self.results["forces"] = f.squeeze
        self.results["virial"] = [x.reshape(3, 3) for x in v]

        # convert virial into stress for lattice relaxation
        if "stress" in properties:
            self.results["stress"] = []
            for i, atoms in enumerate(frames):
                if sum(atoms.get_pbc()) > 0:
                    # the usual convention (tensile stress is positive)
                    # stress = -virial / volume
                    stress = -0.5 * (v[i].copy() + v[i].copy().T) / atoms.get_volume()
                    # Voigt notation
                    self.results["stress"].append(stress.flat[[0, 4, 8, 5, 2, 1]])
                else:
                    raise PropertyNotImplementedError
        
        return
    
    def _convert_from_ase_to_dp(self, atoms):
        """"""
        coord = atoms.get_positions().reshape([1, -1])
        if sum(atoms.get_pbc()) > 0:
            cell = atoms.get_cell(complete=True).reshape([1, -1])
        else:
            cell = None
        symbols = atoms.get_chemical_symbols()
        atype = [self.type_dict[k] for k in symbols]

        return coord, cell, atype


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*

from typing import NoReturn

from .scheduler import AbstractScheduler


class LocalScheduler(AbstractScheduler):
    """Local scheduler."""

    name: str = "local"

    @AbstractScheduler.job_name.setter
    def job_name(self, job_name_: str):
        self._job_name = job_name_
        return

    def submit(self) -> NoReturn:
        """No submit is performed."""

        return

    def is_finished(self) -> bool:
        """Check if the job were finished.

        Returns:
            Always return true.

        """

        return True


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*


from ase.calculators.calculator import Calculator, all_properties, all_changes


class DummyCalculator(Calculator):

    name = "dummy"

    def __init__(
        self, restart=None, label="dummy", atoms=None, directory=".", **kwargs
    ):
        super().__init__(
            restart, label=label, atoms=atoms, directory=directory, **kwargs
        )

        return

    def calculate(
        self, atoms=None, properties=all_properties, system_changes=all_changes
    ):
        """"""
        raise NotImplementedError("DummyCalculator is unable to calculate.")


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc 
import copy
import itertools
import pathlib
from typing import Union, List, Callable, NoReturn, Optional

import numpy as np

from ase import Atoms

from ..core.node import AbstractNode
from ..worker.drive import DriverBasedWorker
from ..data.array import AtomsNDArray


"""Define an AbstractSelector that is the base class of any selector.
"""

def save_cache(fpath, data, random_seed: int=None):
    """"""
    header = ("#{:>11s}  {:>8s}  {:>8s}  {:>8s}  "+"{:>12s}"*4+"\n").format(
        *"index confid step natoms ene aene maxfrc score".split()
    )
    footer = f"random_seed {random_seed}"

    content = header
    for x in data:
        content += ("{:>12s}  {:>8d}  {:>8d}  {:>8d}  "+"{:>12.4f}"*4+"\n").format(*x)
    content += footer

    with open(fpath, "w") as fopen:
        fopen.write(content)

    return

def load_cache(fpath, random_seed: int=None):
    """"""
    with open(fpath, "r") as fopen:
        lines = fopen.readlines()

    # - header
    header = lines[0]

    # - data
    data = lines[1:-1] # TODO: test empty data

    raw_markers = []
    if data:
        # NOTE: new_markers looks like [(0,1),(0,2),(1,0)]
        #       If no structures are selected, the info file should only contain
        #       the header and the footer
        new_markers =[
            [int(x) for x in (d.strip().split()[0]).split(",")] for d in data
        ]
        #new_markers = []
        #for d in data:
        #    curr_marker = []
        #    for x in d.strip().split()[0].split(","):
        #        if x.isdigit(): # MUST BE AN INTEGER
        #            curr_marker.append(int(x))
        #        else:
        #            curr_marker.append(np.nan)
        #    new_markers.append(curr_marker)
        raw_markers = new_markers

    # - footer
    footer = lines[-1]
    cache_random_seed = int(footer.strip().split()[-1]) # TODO: random state
    #assert cache_random_seed == random_seed

    return raw_markers

def group_markers(new_markers_unsorted):
    """"""
    new_markers = sorted(new_markers_unsorted, key=lambda x: x[0])
    raw_markers_unsorted = []
    for k, v in itertools.groupby(new_markers, key=lambda x: x[0]):
        raw_markers_unsorted.append([k,[x[1] for x in v]])

    # traj markers are sorted when set
    raw_markers = [[x[0],sorted(x[1])] for x in sorted(raw_markers_unsorted, key=lambda x:x[0])]

    return raw_markers


class AbstractSelector(AbstractNode):

    """The base class of any selector."""

    #: Selector name.
    name: str = "abstract"

    #: Target axis to select.
    axis: Optional[int] = None

    #: Default parameters.
    default_parameters: dict = dict(
        number = [4, 0.2], # number & ratio
        verbose = False
    )

    #: A worker for potential computations.
    worker: DriverBasedWorker = None

    #: Distinguish structures when using ComposedSelector.
    prefix: str = "selection"

    #: Output file name.
    _fname: str = "info.txt"

    #: Output data format (frames or trajectories).
    _out_fmt: str = "stru"

    def __init__(self, directory="./", axis=None, *args, **kwargs) -> None:
        """Create a selector.

        Args:
            directory: Working directory.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        """
        super().__init__(directory=directory, *args, **kwargs)

        self.axis = axis

        self.fname = self.name+"-info.txt"
        
        # - update parameters from kwargs
        self.parameters = copy.deepcopy(self.default_parameters)
        for k in self.parameters:
            if k in kwargs.keys():
                self.parameters[k] = kwargs[k]

        return

    def set(self, *args, **kwargs):
        """Set parameters."""
        for k, v in kwargs.items():
            if k in self.parameters:
                self.parameters[k] = v

        return

    def __getattr__(self, key):
        """Corresponding getattribute-function."""
        if key != "parameters" and key in self.parameters:
            return self.parameters[key]

        return object.__getattribute__(self, key)

    @AbstractNode.directory.setter
    def directory(self, directory_) -> NoReturn:
        self._directory = pathlib.Path(directory_)
        self.info_fpath = self._directory/self._fname

        return 
    
    @property
    def fname(self):
        """"""
        return self._fname
    
    @fname.setter
    def fname(self, fname_):
        """"""
        self._fname = fname_
        self.info_fpath = self._directory/self._fname
        return
    
    def attach_worker(self, worker=None) -> NoReturn:
        """Attach a worker to this node."""
        self.worker = worker

        return

    def select(self, inp_dat: AtomsNDArray, *args, **kargs) -> List[Atoms]:
        """Select trajectories.

        Based on used selction protocol

        Args:
            frames: A list of ase.Atoms or a list of List[ase.Atoms].
            index_map: Global indices of frames.
            ret_indices: Whether return selected indices or frames.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        
        Returns:
            List[Atoms] or List[int]: selected results

        """
        self._print(f"@@@{self.__class__.__name__}")

        if not self.directory.exists():
            self.directory.mkdir(parents=True)

        # NOTE: input structures should always be the AtomsNDArray type
        # TODO: if inp_dat is already a AtomsNDArray, a new object is created,
        #       which makes markers immutable... Need fix this!!
        if isinstance(inp_dat, AtomsNDArray):
            ...
        else:
            inp_dat = AtomsNDArray(inp_dat)

        frames = inp_dat
        inp_nframes = len(frames.markers)

        # - check if it is finished
        if not (self.info_fpath).exists():
            # -- mark structures
            self._print("run selection...")
            self._mark_structures(frames)
            # -- save cached results for restart
            self._write_cached_results(frames)
        else:
            # -- restart
            self._print("use cached...")
            raw_markers = load_cache(self.info_fpath)
            #print(f"raw_markers: {raw_markers}")
            frames.markers = raw_markers
        
        out_nframes = len(frames.markers)
        self._print(f"{self.name} nstructures {inp_nframes} -> nselected {out_nframes}")

        # - add history
        #   get_marked_structures return reference to Atoms objects
        #for a in inp_dat.get_marked_structures():
        #    print(a.info.get("selection"))

        marked_structures = inp_dat.get_marked_structures()
        for atoms in marked_structures:
            selection = atoms.info.get("selection", "")
            atoms.info["selection"] = selection+f"->{self.name}"
        
        #for a in inp_dat.get_marked_structures():
        #    print(a.info["selection"])

        return marked_structures
    
    @abc.abstractmethod
    def _mark_structures(self, data, *args, **kwargs) -> None:
        """Mark structures subject to selector's conditions."""

        return

    def _parse_selection_number(self, nframes: int) -> int:
        """Compute number of selection based on the input number.

        Args:
            nframes: Number of input frames, sometimes maybe zero.

        """
        default_number, default_ratio = self.default_parameters["number"]
        number_info = self.parameters["number"]
        if isinstance(number_info, int):
            num_fixed, num_percent = number_info, default_ratio
        elif isinstance(number_info, float):
            num_fixed, num_percent = default_number, number_info
        else:
            assert len(number_info) == 2, "Cant parse number for selection..."
            num_fixed, num_percent = number_info
        
        if num_fixed is not None:
            if num_fixed > nframes:
                num_fixed = int(nframes*num_percent)
        else:
            num_fixed = int(nframes*num_percent)

        return num_fixed

    def _write_cached_results(self, aa: AtomsNDArray, *args, **kwargs) -> None:
        """Write selection results into file that can be used for restart."""
        # - 
        markers = aa.markers

        # - output
        data = []
        for ind in markers:
            atoms = aa[tuple(ind)]
            # - gather info
            confid = atoms.info.get("confid", -1)
            step = atoms.info.get("step", -1) # step number in the trajectory
            natoms = len(atoms)
            try:
                ene = atoms.get_potential_energy()
                ae = ene / natoms
            except:
                ene, ae = np.NaN, np.NaN
            try:
                maxforce = np.max(np.fabs(atoms.get_forces(apply_constraint=True)))
            except:
                maxforce = np.NaN
            score = atoms.info.get("score", np.nan)
            # - add info
            ind_str = ",".join([str(x) for x in ind])
            data.append([f"{ind_str}", confid, step, natoms, ene, ae, maxforce, score])
        
        if data:
            save_cache(self.info_fpath, data, self.random_seed)
        else:
            #np.savetxt(
            #    self.info_fpath, [[np.NaN]*8],
            #    header="{:>11s}  {:>8s}  {:>8s}  {:>8s}  {:>12s}  {:>12s}  {:>12s}  {:>12s}".format(
            #        *"index confid step natoms ene aene maxfrc score".split()
            #    ),
            #    footer=f"random_seed {self.random_seed}"
            #)
            content ="{:>11s}  {:>8s}  {:>8s}  {:>8s}  {:>12s}  {:>12s}  {:>12s}  {:>12s}".format(
                *"index confid step natoms ene aene maxfrc score".split()
            )
            content += f"random_seed {self.random_seed}"
            with open(self.info_fpath, "w") as fopen:
                fopen.write(content)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import copy
import pathlib
from typing import List, Callable

from ase import Atoms
from ase.io import read, write

from .. import config
from ..core.node import AbstractNode
from ..data.array import AtomsNDArray
from ..utils.command import dict2str


"""
"""


class StructureBuilder(AbstractNode):

    name = "builder"

    #: Standard print function.
    _print: Callable = config._print

    #: Standard debug function.
    _debug: Callable = config._debug

    def __init__(
        self, use_tags=False, directory="./", random_seed=None, *args, **kwargs
    ):
        """"""
        super().__init__(directory=directory, random_seed=random_seed)

        self.use_tags = use_tags

        return

    @abc.abstractmethod
    def run(self, substrates=None, *args, **kwargs) -> List[Atoms]:
        """Generate structures based on rules."""
        self._print(f"@@@{self.__class__.__name__}")

        self._print(f"RANDOM_SEED : {self.random_seed}")
        rng_state = self.rng.bit_generator.state
        for l in dict2str(rng_state).split("\n"):
            config._print(l)

        if not self.directory.exists():
            self.directory.mkdir(parents=True)

        return


class StructureModifier(StructureBuilder):

    name = "modifier"

    def __init__(self, substrates=None, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

        # TODO: substrates should also be a Builder Object
        # TODO: if substrates is a ChemiclFormula?
        if isinstance(substrates, str) or isinstance(substrates, pathlib.Path):
            substrates = pathlib.Path(substrates).absolute()
        else:
            ...
        # self._print(f"{substrates = }")

        self.substrates = self._load_substrates(substrates)
        # self._print(f"{self.substrates = }")

        return

    def _load_substrates(self, inp_sub) -> List[Atoms]:
        """"""
        substrates = None
        if isinstance(inp_sub, Atoms):
            substrates = [inp_sub]
        elif isinstance(inp_sub, list):  # assume this is a List of Atoms
            substrates = inp_sub
        elif isinstance(inp_sub, AtomsNDArray):
            substrates = inp_sub.get_marked_structures()
        else:
            # assume this is a path
            if isinstance(inp_sub, str) or isinstance(inp_sub, pathlib.Path):
                substrates = read(inp_sub, ":")
            else:
                ...

        return substrates

    def run(self, substrates=None, *args, **kwargs) -> List[Atoms]:
        """"""
        super().run(*args, **kwargs)

        # - load substrates at run
        substrates_at_run = self._load_substrates(substrates)
        if substrates_at_run is not None:
            self.substrates = substrates_at_run

        # TODO: ASE startgenerator mix builders and modifiers
        # assert self.substrates is not None, "Substrates are not set neither at inp nor at run."
        # self.substrates = [copy.deepcopy(s) for s in self.substrates]

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import pathlib
from typing import List

import numpy as np

from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator
from ase.io import read, write
from ase.ga.population import Population
from ase.ga.data import DataConnection

#: Retained keys in key_value_pairs when get_atoms from the database.
RETAINED_KEYS: List[str] = ["extinct", "origin"]


def clean_seed_structures(prev_frames: List[Atoms]):
    """"""
    curr_frames = []
    energies, forces = [], []
    for i, prev_atoms in enumerate(prev_frames):
        # - copy geometry
        curr_atoms = Atoms(
            symbols=copy.deepcopy(prev_atoms.get_chemical_symbols()),
            positions=copy.deepcopy(prev_atoms.get_positions()),
            cell=copy.deepcopy(prev_atoms.get_cell(complete=True)),
            pbc=copy.deepcopy(prev_atoms.get_pbc()),
            tags=prev_atoms.get_tags(),  # retain this for molecules
        )
        # if prev_atoms.get_kinetic_energy() > 0.: # retain this for MD
        #    curr_atoms.set_momenta(prev_atoms.get_momenta())
        curr_frames.append(curr_atoms)

        # - save properties
        try:
            ene = prev_atoms.get_potential_energy()
            energies.append(ene)
        except:
            raise RuntimeError(f"Cannot get energy for seed structure {i}.")

        try:
            frc = prev_atoms.get_forces()
            forces.append(frc)
        except:
            raise RuntimeError(f"Cannot get forces for seed structure {i}.")

    for a, e, f in zip(curr_frames, energies, forces):
        calc = SinglePointCalculator(a, energy=e, forces=f)
        a.calc = calc

    return curr_frames


class AbstractPopulationManager:
    """An abstract population manager for evolutionary algorithms.

    For structure exploration, there are generally two formulations. ASE forms
    current population from all candidates while USPEX forms it based on the
    previous generation. Furthermore, USPEX uses fracGene, fracRand, fracTopRand,
    fracPerm, fracAtomsMut, fracRotMut, fracLatMut, fracSpinMut...

    Example:
        Parameters needed

        $ cat ga.yaml
        population:
            init: # for the initial population
                size: 50 # not necessarily equal to size if set
                seed_file: ./seed.xyz # seed structures for the initial population
            gen: # for the following generations
                size: 20 # number of structures in each generation
                reproduce: 20 # crossover + mutate
                random: 0

    """

    _print = print

    #: Reproduction and mutation.
    MAX_REPROC_TRY: int = 1

    #: Maximum attempts to generate new structures.
    MAX_ATTEMPTS_MULTIPLIER: int = 10

    def __init__(self, params: dict, rng=np.random.default_rng()) -> None:
        """"""
        self.rng = rng

        # - gen params
        gen_params = params.get("gen", dict(size=20, reproduce=20, random=0))
        self.gen_size = gen_params.get("size", None)
        assert isinstance(
            self.gen_size, int
        ), "size of generaton needs to be an integer."

        self.gen_ran_size = gen_params.get("random", 0)
        self.gen_ran_max_try = gen_params.get(
            "max_random_try", self.gen_ran_size * self.MAX_ATTEMPTS_MULTIPLIER
        )

        self.gen_rep_size = gen_params.get("reprod", self.gen_size - self.gen_ran_size)
        self.gen_rep_max_try = gen_params.get(
            "max_reprod_try", self.gen_rep_size * self.MAX_ATTEMPTS_MULTIPLIER
        )

        self.gen_mut_size = gen_params.get(
            "mutate", self.gen_size - self.gen_ran_size - self.gen_rep_size
        )

        # - init params
        init_params = params.get("init", dict(size=20, seed_file=None))
        self.init_size = init_params.get("size", None)
        self.init_seed_file = init_params.get("seed_file", None)

        # - check if params were valid
        assert (
            self.gen_rep_size + self.gen_ran_size + self.gen_mut_size
        ) == self.gen_size, "In each generation, the sum of each component does not equal the total size."
        assert (
            self.gen_ran_size <= self.gen_size
        ), "In each generation, the random size should not be larger than the total size."
        assert (
            self.gen_rep_size <= self.gen_size
        ), "In each generation, the reprod size should not be larger than the total size."
        assert (
            self.gen_mut_size <= self.gen_size
        ), "In each generation, the mutate size should not be larger than the total size."

        # - mutation
        self.pmut = params.get("pmut", 0.5)

        return

    def _get_current_candidates(self, database: DataConnection, curr_gen: int):
        """Get offsprings in the current generation.

        Mutataed candidates do not have `generation` keyword.

        Args:
            database: DataConnection.
            curr_gen: The current generation number.

        """
        candidate_groups = {"paired": [], "random": [], "mutated": []}
        num_paired, num_mutated, num_random = 0, 0, 0

        # unrelaxed_strus_gen_ = list(database.c.select(f"relaxed=0"))
        unrelaxed_strus_gen_ = list(
            database.c.select(f"relaxed=0,generation={curr_gen}")
        )
        for row in unrelaxed_strus_gen_:
            if row.formula:
                # print(row["gaid"], row)
                confid = row["gaid"]
                curr_rows = sorted(
                    database.c.select(f"relaxed=0,gaid={confid}"), key=lambda x: x.mtime
                )
                curr_rows = [x for x in curr_rows if x.formula]
                # - get atoms
                curr_atoms = database.get_atoms(curr_rows[-1].id, add_info=True)
                # NOTE: candidates should not have description info...
                #       otherwise, queued row also has them and failed in
                #       database.c.get_participation_in_pairing()
                kvp = {
                    k: v
                    for k, v in curr_atoms.info["key_value_pairs"].items()
                    if k in RETAINED_KEYS
                }
                data = curr_atoms.info.get(
                    "data", {}
                )  # not every cand has data that stores parents
                curr_atoms.info = {
                    "key_value_pairs": kvp,
                    "data": data,
                    "confid": confid,
                }
                # print(curr_atoms)
                # - count and add atoms
                if "Pairing" in curr_rows[0]["origin"]:
                    num_paired += 1
                    candidate_groups["paired"].append(curr_atoms)
                elif "Mutation" in curr_rows[0]["origin"]:
                    num_mutated += 1
                    candidate_groups["mutated"].append(curr_atoms)
                elif "Random" in curr_rows[0]["origin"]:
                    num_random += 1
                    candidate_groups["random"].append(curr_atoms)
                else:
                    ...

        return candidate_groups, num_paired, num_mutated, num_random

    def _prepare_initial_population(self, generator) -> List[Atoms]:
        self._print("===== Prepare Initial Population =====")
        starting_population = []

        # - try to read seed structures
        # NOTE: seed structures would be re-optimised by the worker
        self._print("----- try to add seed structures -----")
        seed_frames = []
        if self.init_seed_file is not None:
            self._print(str(self.init_seed_file))
            if isinstance(self.init_seed_file, str):
                seed_frames = read(self.init_seed_file, ":")
            elif isinstance(self.init_seed_file, pathlib.Path):
                seed_frames = read(self.init_seed_file, ":")
            elif isinstance(self.init_seed_file, list):  # List[Atoms]
                seed_frames = self.init_seed_file
            else:
                raise RuntimeError(
                    f"Init_seed_file {self.init_seed_file} formst is unsuppoted."
                )
            seed_frames = clean_seed_structures(seed_frames)
            seed_size = len(seed_frames)
            self._print(f"number of seed frames: {seed_size}")
            assert (
                seed_size > 0 and seed_size <= self.init_size
            ), "The number of seed structures is invalid."
        else:
            seed_size = 0

        # TODO: seed structures will be calculated again??
        # TODO: check atom permutation
        for i, atoms in enumerate(seed_frames):
            # TODO: check atom order
            atoms.info["data"] = {}
            atoms.info["key_value_pairs"] = {}
            atoms.info["key_value_pairs"]["origin"] = "seed {}".format(i)
            atoms.info["key_value_pairs"]["raw_score"] = -atoms.get_potential_energy()
            # TODO: check geometric convergence
        self._print(f"number of seed structures: {len(seed_frames)}")
        starting_population.extend(seed_frames)

        # - generate random structures
        self._print("----- try to generate random structures -----")
        random_frames = generator.run(size=self.init_size - seed_size)
        self._print(f"number of random structures: {len(random_frames)}")
        starting_population.extend(random_frames)

        if len(starting_population) != self.init_size:
            raise RuntimeError(
                "It fails to generate the initial population. Check the seed file and the system setting."
            )

        self._print(f"finished creating initial population...")

        return starting_population

    def _prepare_current_population(
        self,
        database: DataConnection,
        curr_gen,
        population,
        generator,
        pairing,
        mutations,
        candidate_groups: dict = {},
        num_paired=0,
        num_mutated=0,
        num_random=0,
    ):
        """Prepare current population.

        Usually, it should be the same as the initial size. However, for variat
        composition search, a large init size can be useful.

        """
        current_candidates = []

        # - reproduction and then mutation
        rest_rep_size = self.gen_rep_size - num_paired
        paired_structures = []
        paired_structures.extend(candidate_groups.get("paired", []))
        if rest_rep_size > 0 and num_random == 0:
            # pair finished but not enough, random already starts...
            for i in range(self.gen_rep_max_try):
                atoms = self._reproduce(database, population, pairing, mutations)
                if atoms is not None:
                    paired_structures.append(atoms)
                    self._print(f"  --> {atoms.info}")
                    self._print("  --> confid %d\n" % (atoms.info["confid"]))
                if len(paired_structures) == self.gen_rep_size:
                    break
            else:
                self._print(
                    f"There is not enough paired structures after {self.gen_rep_max_try} attempts."
                )
        else:
            ...
        current_candidates.extend(paired_structures)

        # - random
        if len(paired_structures) < self.gen_rep_size:
            self._print("There is not enough reproduced (paired) structures.")
            self._print(
                f"Only {len(paired_structures)} are reproduced. The rest would be generated randomly."
            )
            curr_ran_size = self.gen_size - len(paired_structures) - self.gen_mut_size
        else:
            curr_ran_size = self.gen_ran_size

        rest_ran_size = curr_ran_size - num_random
        gen_ran_max_try = rest_ran_size * self.MAX_ATTEMPTS_MULTIPLIER

        random_structures = []
        random_structures.extend(candidate_groups.get("random", []))
        if rest_ran_size > 0 and num_mutated == 0:
            # random finished but not enough, mutation already starts...
            for i in range(gen_ran_max_try):
                frames = generator.run(size=1, soft_error=True)
                if frames:
                    atoms = frames[0]
                    self._print("  reproduce randomly ")
                    atoms.info["key_value_pairs"] = {
                        "extinct": 0,
                        "origin": "RandomCandidateUnrelaxed",
                    }
                    atoms.info["data"] = {}
                    confid = database.c.write(
                        atoms,
                        origin="RandomCandidateUnrelaxed",
                        relaxed=0,
                        extinct=0,
                        generation=curr_gen,
                        key_value_pairs=atoms.info["key_value_pairs"],
                        data=atoms.info["data"],
                    )
                    database.c.update(confid, gaid=confid)
                    atoms.info["confid"] = confid

                    self._print("  --> confid %d\n" % (atoms.info["confid"]))
                    random_structures.append(atoms)
                if len(random_structures) == curr_ran_size:
                    break
            else:
                if self.gen_ran_size > 0:  # NOTE: no break when random size is 0
                    self._print(
                        f"There is not enough random structures after {self.gen_ran_max_try} attempts."
                    )
        else:
            ...
        current_candidates.extend(random_structures)

        # - mutate
        if len(current_candidates) < (self.gen_rep_size + self.gen_ran_size):
            self._print("There is not enough reproduced+random structures.")
            self._print(
                f"Only {len(current_candidates)} are generated. The rest would be generated by mutations."
            )
            curr_mut_size = self.gen_size - len(current_candidates)
        else:
            curr_mut_size = self.gen_mut_size

        rest_mut_size = curr_mut_size - num_mutated
        gen_mut_max_try = rest_mut_size * self.MAX_ATTEMPTS_MULTIPLIER

        mutated_structures = []
        mutated_structures.extend(candidate_groups.get("mutated", []))
        for i in range(gen_mut_max_try):
            atoms = population.get_one_candidate(with_history=True)
            a3, desc = mutations.get_new_individual([atoms])
            if atoms is not None:
                database.add_unrelaxed_step(a3, desc)
                self._print("  Mutate cand{} by {}".format(atoms.info["confid"], desc))
                self._print("  --> confid %d\n" % (a3.info["confid"]))
                mutated_structures.append(a3)
            if len(mutated_structures) == curr_mut_size:
                break
        else:
            if self.gen_mut_size > 0:  # NOTE: no break when random size is 0
                self._print(
                    f"There is not enough mutated structures after {gen_mut_max_try} attempts."
                )
        current_candidates.extend(mutated_structures)

        if len(current_candidates) != self.gen_size:
            self._print("Not enough candidates for the next generation.")
            raise RuntimeError("Not enough candidates for the next generation.")

        return current_candidates

    def _update_generation_settings(self, population, mutations, pairing):
        """Update some generation-specific settings."""
        # - operations at the end of each generation
        cur_pop = population.get_current_population()
        # find_strain = False
        # from ase.ga.standardmutations import StrainMutation
        for mut in mutations.oplist:
            # if issubclass(mut, StrainMutation):
            #    find_strain = True
            #    mut.update_scaling_volume(cur_pop, w_adapt=0.5, n_adapt=0)
            #    self._print(f"StrainMutation Scaling Volume: {mut.scaling_volume}")
            if hasattr(mut, "update_scaling_volume"):
                mut.update_scaling_volume(cur_pop, w_adapt=0.5, n_adapt=0)
                self._print(
                    f"{mut.__class__.__name__} Scaling Volume: {mut.scaling_volume}"
                )
        if hasattr(pairing, "update_scaling_volume"):
            pairing.update_scaling_volume(cur_pop, w_adapt=0.5, n_adapt=0)
            self._print(
                f"{pairing.__class__.__name__} Scaling Volume: {pairing.scaling_volume}"
            )

        return

    def _reproduce(self, database, population, pairing, mutations) -> Atoms:
        """Reproduce a structure from the current population."""
        a3 = None
        for i in range(self.MAX_REPROC_TRY):
            # try 10 times
            parents = population.get_two_candidates()
            a3, desc = pairing.get_new_individual(
                parents
            )  # NOTE: this also adds key_value_pairs to a.info
            if a3 is not None:
                database.add_unrelaxed_candidate(
                    a3,
                    description=desc,  # here, desc is used to add "pairing": 1 to database
                )  # if mutation happens, it will not be relaxed

                mut_desc = ""
                curr_prob = self.rng.random()
                if curr_prob < self.pmut:
                    a3_mut, mut_desc = mutations.get_new_individual([a3])
                    if a3_mut is not None:
                        self._print(f"a3_mut: {a3_mut.info}")
                        database.add_unrelaxed_step(a3_mut, mut_desc)
                        # write("./pairs.xyz", a3, append=True)
                        # write("./muts.xyz", a3_mut, append=True)
                        a3 = a3_mut
                    else:
                        mut_desc = ""
                self._print(
                    f"  reproduce offspring with {desc} and ({curr_prob}) {mut_desc} after {i+1} attempts..."
                )
                break
            else:
                mut_desc = ""
                self._print(
                    f"  failed to reproduce offspring with {desc} {mut_desc} after {i+1} attempts..."
                )
        else:
            self._print(
                "cannot reproduce offspring a3 after {0} attempts".format(
                    self.MAX_REPROC_TRY
                )
            )

        return a3


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import collections
import itertools
import pathlib

from typing import NoReturn, List, Union

import omegaconf

import numpy as np

from ase import Atoms
from ase.io import read, write

from ..core.placeholder import Placeholder
from ..core.variable import Variable
from ..core.operation import Operation
from ..core.register import registers
from ..data.array import AtomsNDArray
from .builder import StructureBuilder


@registers.placeholder.register
class StructurePlaceholder(Placeholder):

    name = "structure"

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__()

        return


@registers.variable.register
class BuilderVariable(Variable):
    """Build structures from the scratch."""

    def __init__(self, directory="./", *args, **kwargs):
        """"""
        # - create a validator
        method = kwargs.get("method", "direct")
        builder = registers.create("builder", method, convert_name=False, **kwargs)

        super().__init__(initial_value=builder, directory=directory)

        return

    @Variable.directory.setter
    def directory(self, directory_) -> NoReturn:
        """"""
        self._directory = pathlib.Path(directory_)

        # Value is the attached builder
        self.value.directory = self._directory

        return

    def _reset_random_seed(self, mode: str = "init"):
        """Rewind random state to the one at the initialisation of the object."""
        if mode == "init":
            self.value.set_rng(seed=self.value.init_random_seed)
        elif mode == "zero":
            self.value.set_rng(seed=None)
        else:
            raise RuntimeError(f"INCORRECT RESET RANDOM SEED MODE {mode}.")

        return


@registers.operation.register
class read_stru(Operation):

    def __init__(
        self, fname, format=None, index=":", input_nodes=[], directory="./", **kwargs
    ) -> None:
        """"""
        super().__init__(input_nodes, directory)

        self.fname = fname  # This is broadcastable...
        self.format = format
        self.index = index
        self.kwargs = kwargs

        return

    def forward(self, *args, **kwargs) -> AtomsNDArray:
        """"""
        super().forward()

        # - check params
        if isinstance(self.fname, str):
            fname_ = [self.fname]
        else:  # assume it is an iterable object
            fname_ = self.fname

        # - read structures
        frames = []
        for curr_fname in fname_:
            self._print(f"read {curr_fname}")
            curr_frames = read(
                curr_fname, format=self.format, index=self.index, **self.kwargs
            )
            if isinstance(curr_frames, Atoms):
                curr_frames = [
                    curr_frames
                ]  # if index is single, then read will give Atoms
            frames.extend(curr_frames)

        frames = AtomsNDArray(frames)
        self._print(f"shape of structures: {frames.shape}")

        self.status = "finished"

        return frames


@registers.operation.register
class write_stru(Operation):

    def __init__(
        self, fname, structures, format="extxyz", directory="./", *args, **kwargs
    ) -> None:
        """"""
        input_nodes = [structures]
        super().__init__(input_nodes, directory)

        self.fname = fname
        self.format = format
        self.kwargs = kwargs

        return

    def forward(self, structures, *args, **kwargs):
        """"""
        super().forward()

        write(self.directory / self.fname, structures, format=self.format)

        self.status = "finished"

        return


@registers.operation.register
class xbuild(Operation):
    """Build structures without substrate structures."""

    def __init__(self, builders, directory="./") -> NoReturn:
        super().__init__(input_nodes=builders, directory=directory)

    def forward(self, *args, **kwargs) -> List[Atoms]:
        """"""
        super().forward()

        bundle = []
        for i, builder in enumerate(args):
            builder.directory = self.directory
            curr_builder_output = self.directory / f"{builder.name}_output-{i}.xyz"
            if curr_builder_output.exists():
                frames = read(curr_builder_output, ":")
            else:
                frames = builder.run()
                write(curr_builder_output, frames)
            self._print(f"{i} - {builder.name} nframes: {len(frames)}")
            bundle.extend(frames)
        self._print(f"nframes: {len(bundle)}")

        self.status = "finished"

        return bundle


@registers.operation.register
class build(Operation):
    """Build structures without substrate structures."""

    def __init__(self, builder, size: int = 1, directory="./") -> NoReturn:
        super().__init__(input_nodes=[builder], directory=directory)

        self.size = size

        return

    def forward(self, builder) -> List[Atoms]:
        """"""
        super().forward()

        builder.directory = self.directory
        curr_builder_output = self.directory / f"{builder.name}_output.xyz"
        if curr_builder_output.exists():
            frames = read(curr_builder_output, ":")
        else:
            frames = builder.run(size=self.size)
            write(curr_builder_output, frames)
        self._print(f"{builder.name} nframes: {len(frames)}")

        self.status = "finished"

        return frames


@registers.operation.register
class modify(Operation):

    def __init__(
        self, substrates, modifier, size: int = 1, repeat: int = 1, directory="./"
    ) -> None:
        """"""
        super().__init__(input_nodes=[substrates, modifier], directory=directory)

        self.size = size  # create number of new structures
        self.repeat = repeat  # repeat modification times for one structure

        return

    @Operation.directory.setter
    def directory(self, directory_) -> None:
        """"""
        self._directory = pathlib.Path(directory_)

        self.input_nodes[0].directory = self._directory / "substrates"
        self.input_nodes[1].directory = self._directory / "modifier"

        return

    def _preprocess_input_nodes(self, input_nodes):
        """"""
        substrates, modifier = input_nodes
        if isinstance(substrates, str) or isinstance(substrates, pathlib.Path):
            # TODO: check if it is a molecule name
            substrates = build(
                BuilderVariable(
                    directory=self.directory / "substrates",
                    method="reader",
                    fname=substrates,
                )
            )
        if isinstance(modifier, dict) or isinstance(
            modifier, omegaconf.dictconfig.DictConfig
        ):
            modifier = BuilderVariable(
                directory=self.directory / "modifier", **modifier
            )

        return substrates, modifier

    def forward(self, substrates: List[Atoms], modifier) -> List[Atoms]:
        """Modify inputs structures.

        A modifier only accepts one structure each time.

        """
        super().forward()

        cache_path = self.directory / f"{modifier.name}-out.xyz"
        if not cache_path.exists():
            frames = modifier.run(substrates, size=self.size)
            write(cache_path, frames)
        else:
            frames = read(cache_path, ":")

        self.status = "finished"

        return frames


def create_builder(config: Union[str, dict]) -> StructureBuilder:
    """"""
    supported_configtypes = ["json", "yaml"]
    if isinstance(config, (str, pathlib.Path)):
        params = str(config)
        suffix = params[-4:]
        if suffix in supported_configtypes:
            from gdpx.utils.command import parse_input_file

            params = parse_input_file(config)
        else:  # assume it is an ASE readable structure file
            params = dict(method="direct", frames=params)

    builder = BuilderVariable(**params).value

    return builder


def build_structures(config: dict, substrates=None, size: int = 1, directory="./"):
    """"""
    directory = pathlib.Path(directory)

    builder = BuilderVariable(directory=directory, **config).value
    builder.directory = directory

    # assume substrates is a file path
    if substrates is not None:
        substrates = read(substrates, ":")

    frames = builder.run(substrates=substrates, size=size)

    write(directory / "structures.xyz", frames)

    return


if __name__ == "__main__":
    ...



#!/usr3/bin/env python3
# -*- coding: utf-8 -*


import pathlib

from .. import AbstractPotentialManager, DummyCalculator


class PlumedManager(AbstractPotentialManager):

    name = "plumed"

    implemented_backends = ["ase"]

    valid_combinations = (
        # calculator, dynamics
        ("ase", "ase"),
    )

    def __init__(self) -> None:
        """"""

        return

    def register_calculator(self, calc_params: dict, *agrs, **kwargs) -> None:
        """"""
        super().register_calculator(calc_params, *agrs, **kwargs)

        calc = DummyCalculator()
        if self.calc_backend == "ase":
            try:
                from .calculators.plumed2 import Plumed
            except:
                raise ModuleNotFoundError(
                    "Please install py-plumed to use the ase interface."
                )

            inp = calc_params.get("inp", "./plumed.inp")
            if isinstance(inp, str) or isinstance(inp, pathlib.Path):
                inp = pathlib.Path(inp)
                if inp.exists():
                    input_lines = []
                    with open(inp, "r") as fopen:
                        lines = fopen.readlines()
                        for line in lines:
                            line = line.strip()
                            if line and not line.startswith("#"):
                                if "#" in line:
                                    line = line[: line.index("#")]
                                else:
                                    line = line
                                input_lines.append(line + "\n")
                    self.calc_params.update(inp=input_lines)
                else:
                    raise FileNotFoundError(f"{inp} does not exist.")
            elif isinstance(inp, list):
                input_lines = inp
            else:
                raise ValueError(f"Plumed input {inp} {type(inp)} is invalid.")

            kT = calc_params.get("kT", 1.0)
            use_charge = calc_params.get("use_charge", False)
            update_charge = calc_params.get("update_charge", False)
            calc = Plumed(
                input=input_lines,
                kT=kT,
                use_charge=use_charge,
                update_charge=update_charge,
            )
        else:
            ...

        self.calc = calc

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import argparse
import dataclasses
from typing import List
import pathlib
import traceback
import warnings

import numpy as np

from ase import units
from ase import Atoms
from ase.io import read, write
from ase.calculators.calculator import FileIOCalculator
from ase.calculators.singlepoint import SinglePointCalculator
from ase.calculators.cp2k import CP2K, parse_input, InputSection
from ase.constraints import FixAtoms

from .driver import DriverSetting, AbstractDriver
from ..builder.constraints import parse_constraint_info

"""Convert cp2k md outputs to ase xyz file.
"""

UNCONVERGED_SCF_FLAG: str = "*** WARNING in qs_scf.F:598 :: SCF run NOT converged ***"
ABORT_FLAG: str = "ABORT"

def read_cp2k_xyz(fpath):
    """Read xyz-like file by cp2k.

    Accept prefix-pos-1.xyz or prefix-frc-1.xyz.
    """
    # - read properties
    frame_steps, frame_times, frame_energies = [], [], []
    frame_symbols = []
    frame_properties = [] # coordinates or forces
    with open(fpath, "r") as fopen:
        while True:
            line = fopen.readline()
            if not line:
                break
            natoms = int(line.strip().split()[0])
            symbols, properties = [], []
            line = fopen.readline() # energy line
            info_data = line.strip().split()
            frame_energies.append(info_data[-1])
            for i in range(natoms):
                line = fopen.readline()
                data_line = line.strip().split()
                symbols.append(data_line[0])
                properties.append(data_line[1:])
            frame_symbols.append(symbols)
            frame_properties.append(properties)

    return frame_symbols, frame_energies, frame_properties


def read_cp2k_outputs(wdir, prefix: str="cp2k") -> List[Atoms]:
    """"""
    wdir = pathlib.Path(wdir)

    # - positions
    pos_fpath = wdir / (prefix+"-pos-1.xyz")
    frame_symbols, frame_energies, frame_positions = read_cp2k_xyz(pos_fpath)
    # NOTE: cp2k uses a.u. and we use eV
    frame_energies = np.array(frame_energies, dtype=np.float64)
    frame_energies *= units.Hartree # 2.72113838565563E+01
    # NOTE: cp2k uses AA the same as we do
    frame_positions = np.array(frame_positions, dtype=np.float64)
    #frame_positions *= 5.29177208590000E-01
    #print("shape of positions: ", frame_positions.shape)

    # - forces
    frc_fpath = wdir / (prefix+"-frc-1.xyz")
    _, _, frame_forces = read_cp2k_xyz(frc_fpath)
    # NOTE: cp2k uses a.u. and we use eV/AA
    frame_forces = np.array(frame_forces, dtype=np.float64)
    frame_forces *= units.Hartree/units.Bohr #(2.72113838565563E+01/5.29177208590000E-01)
    #print("shape of forces: ", frame_forces.shape)

    # - simulation box
    # TODO: parse cell from inp or out
    box_fpath = wdir / (prefix+"-1.cell")
    with open(box_fpath, "r") as fopen:
        # Step   Time [fs]       
        # Ax [Angstrom]       Ay [Angstrom]       Az [Angstrom]       
        # Bx [Angstrom]       By [Angstrom]       Bz [Angstrom]       
        # Cx [Angstrom]       Cy [Angstrom]       Cz [Angstrom]      Volume [Angstrom^3]
        lines = fopen.readlines()
        data = np.array(
            [line.strip().split() for line in lines[1:]], dtype=np.float64
        )
    steps = data[:, 0]
    boxes = data[:, 2:-1]

    # TODO: step must be int?
    # attach forces to frames, zip the shortest
    frames = []
    for step, symbols, box, positions, energy, forces in zip(
        steps, frame_symbols, boxes, frame_positions, frame_energies, frame_forces
    ):
        atoms = Atoms(
            symbols, positions=positions,
            cell=box.reshape(3,3), 
            pbc=[1,1,1] # TODO: should determine in the cp2k input file
        )
        atoms.info["step"] = int(step)
        spc = SinglePointCalculator(
            atoms=atoms, energy=energy, 
            free_energy=energy, # TODO: depand on electronic method used
            forces=forces
        )
        atoms.calc = spc
        frames.append(atoms)

    return frames

@dataclasses.dataclass
class Cp2kDriverSetting(DriverSetting):

    fmax: float = 4.5e-4*(units.Hartree/units.Bohr)

    def __post_init__(self):
        """"""
        pairs = [] # key-value pairs that avoid conflicts by same keys
        if self.task == "min":
            # - fmax and steps can be set on-the-fly, see get_run_params
            method = self.min_style.upper()
            pairs.extend(
                [
                    ("GLOBAL", "RUN_TYPE GEO_OPT"),
                    ("MOTION/GEO_OPT", "TYPE MINIMIZATION"),
                    #("MOTION/GEO_OPT", f"MAX_ITER {self.steps}"),
                    ("MOTION/GEO_OPT", f"OPTIMIZER {method}"),
                    ("MOTION/PRINT/RESTART_HISTORY/EACH", f"GEO_OPT {self.ckpt_period}"),
                ]
            )
            #if self.fmax is not None:
            #    pairs.append(
            #        ("MOTION/GEO_OPT", f"MAX_FORCE {self.fmax/(units.Hartree/units.Bohr)}")
            #    )
        
        if self.task == "md":
            # - steps can be set on-the-fly, see get_run_params
            md_style = self.md_style.upper()
            pairs.extend(
                [
                    ("GLOBAL", "RUN_TYPE MD"),
                    ("MOTION/MD", f"TIMESTEP {self.timestep}"),
                    ("MOTION/PRINT/RESTART_HISTORY/EACH", f"MD {self.ckpt_period}"),
                ]
            )
            if self.md_style == "nve":
                pairs.extend(
                    [
                        ("MOTION/MD", "ENSEMBLE NVE"),
                        ("MOTION/MD", f"TEMPERATURE {self.temp}"),
                    ]
                )
            elif self.md_style == "nvt":
                # TODO: support different thermostats...
                pairs.extend(
                    [
                        ("MOTION/MD", "ENSEMBLE NVT"),
                        ("MOTION/MD", f"TEMPERATURE {self.temp}"),
                        #("MOTION/MD/THERMOSTAT/NOSE/TIMECON", "13.34")
                        ("MOTION/MD/THERMOSTAT/CSVR", f"TIMECON {self.Tdamp}")
                    ]
                )
            elif self.md_style == "npt":
                pairs.extend(
                    [
                        ("MOTION/MD", "ENSEMBLE NPT_I"),
                        ("MOTION/MD", f"TEMPERATURE {self.temp}"),
                        ("MOTION/MD/THERMOSTAT/NOSE", f"TIMECON {self.Tdamp}"),
                        ("MOTION/MD/BAROSTAT", f"PRESSURE {self.press}"),
                        ("MOTION/MD/BAROSTAT", f"TIMECON {self.Pdamp}"),
                    ]
                )
            else:
                ...
        
        # TODO: Move to manager? as we always need this outputs to convert calculation
        #       to a trajectory
        pairs.extend(
            [
                ("MOTION/PRINT/CELL", "_SECTION_PARAMETERS_ ON"),
                ("MOTION/PRINT/TRAJECTORY", "_SECTION_PARAMETERS_ ON"),
                ("MOTION/PRINT/FORCES", "_SECTION_PARAMETERS_ ON"),
            ]
        )
        self._internals["pairs"] = pairs

        return
    
    def get_run_params(self, *args, **kwargs):
        """"""
        # - convergence criteria
        fmax_ = kwargs.get("fmax", self.fmax)
        steps_ = kwargs.get("steps", self.steps)

        run_pairs = []
        if self.task == "min":
            run_pairs.append(
                ("MOTION/GEO_OPT", f"MAX_ITER {steps_}"),
            )
            if fmax_ is not None:
                run_pairs.append(
                    ("MOTION/GEO_OPT", f"MAX_FORCE {fmax_/(units.Hartree/units.Bohr)}")
                )
        if self.task == "md":
            run_pairs.append(
                ("MOTION/MD", f"STEPS {steps_}"),
            )

        # - add constraint
        run_params = dict(
            constraint = kwargs.get("constraint", self.constraint),
            run_pairs = run_pairs
        )

        return run_params

class Cp2kDriver(AbstractDriver):

    name = "cp2k"

    default_task = "min"
    supported_tasks = ["min", "md"]

    #saved_fnames = [
    #    "cp2k.inp", "cp2k.out", "cp2k-1.cell", "cp2k-pos-1.xyz", "cp2k-frc-1.xyz",
    #    "cp2k-BFGS.Hessian"
    #]

    def __init__(self, calc, params: dict, directory="./", *args, **kwargs):
        """"""
        super().__init__(calc, params, directory, *args, **kwargs)

        self.setting = Cp2kDriverSetting(**params)

        return
    
    def _verify_checkpoint(self, *args, **kwargs) -> bool:
        """"""
        verified = super()._verify_checkpoint(*args, **kwargs)
        if verified:
            checkpoints = list(self.directory.glob("*.restart"))
            self._debug(f"checkpoints: {checkpoints}")
            if not checkpoints:
                verified = False
        else:
            ...

        return verified
    
    def _irun(self, atoms: Atoms, ckpt_wdir=None, *args, **kwargs):
        """"""
        try:
            if ckpt_wdir is None: # start from the scratch
                run_params = self.setting.get_run_params(**kwargs)
                run_params.update(**self.setting.get_init_params())

                # - update input template
                # GLOBAL section is automatically created...
                # FORCE_EVAL.(METHOD, POISSON)
                inp = self.calc.parameters.inp # string
                sec = parse_input(inp)
                for (k, v) in run_params["pairs"]:
                    sec.add_keyword(k, v)
                for (k, v) in run_params["run_pairs"]:
                    sec.add_keyword(k, v)

                # -- check constraint
                cons_text = run_params.pop("constraint", None)
                mobile_indices, frozen_indices = parse_constraint_info(atoms, cons_text, ret_text=False)
                if frozen_indices:
                    #atoms._del_constraints()
                    #atoms.set_constraint(FixAtoms(indices=frozen_indices))
                    frozen_indices = sorted(frozen_indices)
                    sec.add_keyword(
                        "MOTION/CONSTRAINT/FIXED_ATOMS", 
                        "LIST {}".format(" ".join([str(i+1) for i in frozen_indices]))
                    )
            else:
                with open(ckpt_wdir/"cp2k.inp", "r") as fopen:
                    inp = "".join(fopen.readlines())
                sec = parse_input(inp)

                def remove_keyword(section, keywords):
                    """"""
                    for kw in keywords:
                        parts = kw.upper().split("/")
                        subsection = section.get_subsection("/".join(parts[0:-1]))
                        new_subkeywords = []
                        for subkw in subsection.keywords:
                            if parts[-1] not in subkw:
                                new_subkeywords.append(subkw)
                        subsection.keywords = new_subkeywords

                    return
                
                remove_keyword(
                    sec, 
                    keywords=[ # avoid conflicts
                        "GLOBAL/PROJECT", "GLOBAL/PRINT_LEVEL", "FORCE_EVAL/METHOD", 
                        "FORCE_EVAL/DFT/BASIS_SET_FILE_NAME", "FORCE_EVAL/DFT/POTENTIAL_FILE_NAME",
                        "FORCE_EVAL/SUBSYS/CELL/PERIODIC", 
                        "FORCE_EVAL/SUBSYS/CELL/A", "FORCE_EVAL/SUBSYS/CELL/B", "FORCE_EVAL/SUBSYS/CELL/C"
                    ]
                )

                sec.add_keyword("EXT_RESTART/RESTART_FILE_NAME", str(ckpt_wdir/"cp2k-1.restart"))
                sec.add_keyword("FORCE_EVAL/DFT/SCF/SCF_GUESS", "RESTART")

                # - copy wavefunctions...
                restart_wfns = sorted(list(ckpt_wdir.glob("*.wfn")))
                for wfn in restart_wfns:
                    (self.directory/wfn.name).symlink_to(wfn, target_is_directory=False)

            self.calc.parameters.inp = "\n".join(sec.write())
            atoms.calc = self.calc

            _ = atoms.get_forces()

        except Exception as e:
            self._debug(e)
            self._debug(traceback.print_exc())

        return
    
    def _read_a_single_trajectory(self, wdir, *args, **kwargs):
        """"""
        frames = read_cp2k_outputs(wdir, prefix=self.name)

        return frames
    
    def read_trajectory(self, *args, **kwargs) -> List[Atoms]:
        """"""
        super().read_trajectory(*args, **kwargs)

        # TODO: support restart!!!
        try:
            # -- read backups
            prev_wdirs = sorted(self.directory.glob(r"[0-9][0-9][0-9][0-9][.]run"))
            self._debug(f"prev_wdirs: {prev_wdirs}")

            traj_list = []
            for w in prev_wdirs:
                curr_frames = self._read_a_single_trajectory(w)
                traj_list.append(curr_frames)
            
            cp2ktraj = self.directory / "cp2k-pos-1.xyz"
            if cp2ktraj.exists() and cp2ktraj.stat().st_size != 0:
                traj_list.append(read_cp2k_outputs(self.directory, prefix=self.name))

            # -- concatenate
            traj_frames, ntrajs = [], len(traj_list)
            if ntrajs > 0:
                traj_frames.extend(traj_list[0])
                for i in range(1, ntrajs):
                    assert np.allclose(traj_list[i-1][-1].positions, traj_list[i][0].positions), f"Traj {i-1} and traj {i} are not consecutive."
                    traj_frames.extend(traj_list[i][1:])
            else:
                ...
        except Exception as e:
            self._debug(e)
            self._debug(traceback.print_exc())

        return traj_frames 

    
class Cp2kFileIO(FileIOCalculator):

    implemented_properties = ["energy", "free_energy", "forces", "stress"]
    command = None

    default_parameters = dict(
        auto_write=False,
        basis_set="DZVP-MOLOPT-SR-GTH",
        basis_set_file="BASIS_MOLOPT",
        pseudo_potential="GTH-PBE",
        potential_file="POTENTIAL",
        inp="",
        force_eval_method="Quickstep",
        charge=0,
        uks=False,
        stress_tensor=False,
        poisson_solver="auto",
        xc="PBE",
        max_scf=50,
        cutoff=400 * units.Rydberg,
        print_level="MEDIUM"
    )

    """This calculator is consistent with v9.1 and v2022.1.
    """

    def __init__(self, restart=None, label="cp2k", atoms=None, command="cp2k.psmp", **kwargs):
        """Construct CP2K-calculator object"""
        super().__init__(restart=restart, label=label, atoms=atoms, command=command, **kwargs)

        # complete command
        command_ = self.command
        if "-i" in command_:
            ...
        else:
            label_name = pathlib.Path(self.label).name
            command_ += f" -i {label_name}.inp -o {label_name}.out"
        self.command = command_

        return

    def read_results(self):
        """"""
        super().read_results()

        label_name = pathlib.Path(self.label).name

        # check run_type
        #run_type = "md"
        #if run_type.upper() == "MD":
        #    trajectory = read_cp2k_outputs(self.directory, prefix=label_name)
        #else:
        #    ... # GEO_OPT, CELL_OPT
        trajectory = read_cp2k_outputs(self.directory, prefix=label_name)
        
        atoms = trajectory[-1]
        self.results["energy"] = atoms.get_potential_energy()
        self.results["free_energy"] = atoms.get_potential_energy(force_consistent=True)
        self.results["forces"] = atoms.get_forces()
        # TODO: stress
        
        scf_convergence = self.read_convergence()
        atoms.info["scf_convergence"] = scf_convergence
        if not scf_convergence:
            atoms.info["error"] = f"Unconverged SCF at {self.directory}."

        return
    
    def write_input(self, atoms, properties=None, system_changes=None):
        """"""
        super().write_input(atoms, properties, system_changes)

        label_name = pathlib.Path(self.label).name
        wdir = pathlib.Path(self.directory)
        with open(wdir/f"{label_name}.inp", "w") as fopen:
            fopen.write(self._generate_input())

        return
    
    def _generate_input(self):
        """Generates a CP2K input file"""
        p = self.parameters
        root = parse_input(p.inp)
        label_name = pathlib.Path(self.label).name
        root.add_keyword('GLOBAL', 'PROJECT ' + label_name)
        if p.print_level:
            root.add_keyword('GLOBAL', 'PRINT_LEVEL ' + p.print_level)
        #root.add_keyword("GLOBAL", "RUN_TYPE " + "CELL_OPT")
        if p.force_eval_method:
            root.add_keyword('FORCE_EVAL', 'METHOD ' + p.force_eval_method)
        if p.stress_tensor:
            root.add_keyword('FORCE_EVAL', 'STRESS_TENSOR ANALYTICAL')
            root.add_keyword('FORCE_EVAL/PRINT/STRESS_TENSOR',
                             '_SECTION_PARAMETERS_ ON')
        if p.basis_set_file:
            root.add_keyword('FORCE_EVAL/DFT',
                             'BASIS_SET_FILE_NAME ' + p.basis_set_file)
        if p.potential_file:
            root.add_keyword('FORCE_EVAL/DFT',
                             'POTENTIAL_FILE_NAME ' + p.potential_file)
        if p.cutoff:
            root.add_keyword('FORCE_EVAL/DFT/MGRID',
                             'CUTOFF [eV] %.18e' % p.cutoff)
        if p.max_scf:
            root.add_keyword('FORCE_EVAL/DFT/SCF', 'MAX_SCF %d' % p.max_scf)
            root.add_keyword('FORCE_EVAL/DFT/LS_SCF', 'MAX_SCF %d' % p.max_scf)

        if p.xc:
            legacy_libxc = ""
            for functional in p.xc.split():
                functional = functional.replace("LDA", "PADE")  # resolve alias
                xc_sec = root.get_subsection('FORCE_EVAL/DFT/XC/XC_FUNCTIONAL')
                # libxc input section changed over time
                if functional.startswith("XC_") and self._shell.version < 3.0:
                    legacy_libxc += " " + functional  # handled later
                elif functional.startswith("XC_") and self._shell.version < 5.0:
                    s = InputSection(name='LIBXC')
                    s.keywords.append('FUNCTIONAL ' + functional)
                    xc_sec.subsections.append(s)
                elif functional.startswith("XC_"):
                    s = InputSection(name=functional[3:])
                    xc_sec.subsections.append(s)
                else:
                    s = InputSection(name=functional.upper())
                    xc_sec.subsections.append(s)
            if legacy_libxc:
                root.add_keyword('FORCE_EVAL/DFT/XC/XC_FUNCTIONAL/LIBXC',
                                 'FUNCTIONAL ' + legacy_libxc)

        if p.uks:
            root.add_keyword('FORCE_EVAL/DFT', 'UNRESTRICTED_KOHN_SHAM ON')

        if p.charge and p.charge != 0:
            root.add_keyword('FORCE_EVAL/DFT', 'CHARGE %d' % p.charge)

        # add Poisson solver if needed
        if p.poisson_solver == 'auto' and not any(self.atoms.get_pbc()):
            root.add_keyword('FORCE_EVAL/DFT/POISSON', 'PERIODIC NONE')
            root.add_keyword('FORCE_EVAL/DFT/POISSON', 'PSOLVER  MT')

        # write coords
        syms = self.atoms.get_chemical_symbols()
        atoms = self.atoms.get_positions()
        for elm, pos in zip(syms, atoms):
            line = '%s %.18e %.18e %.18e' % (elm, pos[0], pos[1], pos[2])
            root.add_keyword('FORCE_EVAL/SUBSYS/COORD', line, unique=False)

        # write cell
        pbc = ''.join([a for a, b in zip('XYZ', self.atoms.get_pbc()) if b])
        if len(pbc) == 0:
            pbc = 'NONE'
        root.add_keyword('FORCE_EVAL/SUBSYS/CELL', 'PERIODIC ' + pbc)
        c = self.atoms.get_cell()
        for i, a in enumerate('ABC'):
            line = '%s %.18e %.18e %.18e' % (a, c[i, 0], c[i, 1], c[i, 2])
            root.add_keyword('FORCE_EVAL/SUBSYS/CELL', line)

        # determine pseudo-potential
        potential = p.pseudo_potential
        if p.pseudo_potential == 'auto':
            if p.xc and p.xc.upper() in ('LDA', 'PADE', 'BP', 'BLYP', 'PBE',):
                potential = 'GTH-' + p.xc.upper()
            else:
                msg = 'No matching pseudo potential found, using GTH-PBE'
                warnings.warn(msg, RuntimeWarning)
                potential = 'GTH-PBE'  # fall back

        # write atomic kinds
        subsys = root.get_subsection('FORCE_EVAL/SUBSYS').subsections
        kinds = dict([(s.params, s) for s in subsys if s.name == "KIND"])
        for elem in set(self.atoms.get_chemical_symbols()):
            if elem not in kinds.keys():
                s = InputSection(name='KIND', params=elem)
                subsys.append(s)
                kinds[elem] = s
            if p.basis_set:
                kinds[elem].keywords.append('BASIS_SET ' + p.basis_set)
            if potential:
                kinds[elem].keywords.append('POTENTIAL ' + potential)

        output_lines = ['!!! Generated by ASE !!!'] + root.write()
        return '\n'.join(output_lines)
    
    def read_convergence(self):
        """Read SCF convergence."""
        label_name = pathlib.Path(self.label).name
        cp2kout = pathlib.Path(self.directory) / f"{label_name}.out"

        converged = True
        with open(cp2kout, "r") as fopen:
            while True:
                line = fopen.readline()
                if not line:
                    break
                if line.strip() == UNCONVERGED_SCF_FLAG:
                    converged = False
                    break
                if ABORT_FLAG in line:
                    converged = False
                    break

        return converged


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

""" worker for training potentials
"""

import uuid
import pathlib
from typing import NoReturn, Callable
import warnings
import yaml

import numpy as np

from tinydb import Query, TinyDB

from ..data.interface import DatasetVariable
from .worker import AbstractWorker
from ..potential.trainer import AbstractTrainer


class TrainerBasedWorker(AbstractWorker):

    TRAIN_PREFIX: str = "m"

    #: Whether share a dataset when training a group of models.
    _share_dataset: bool = False

    def __init__(
        self,
        trainer: AbstractTrainer,
        scheduler,
        share_dataset: bool = False,
        auto_submit: bool = True,
        directory=None,
        *args,
        **kwargs,
    ) -> None:
        """Initialise a TrainerBasedWorker.

        Args:
            share_dataset: Whether a group of models are traiend on a shared dataset.
            auto_submit:
                Whether submit scheduler jobs automatically. Otherwise, it must be
                submitted manully.

        """
        super().__init__(directory)

        self.trainer = trainer
        self.scheduler = scheduler

        self._share_dataset = share_dataset
        self._submit = auto_submit

        return

    def _get_train_params(
        self,
        trainer: AbstractTrainer,
        dataset,
        init_model,
        use_shared_dataset: bool = False,
    ) -> dict:
        """"""
        trainer_params = {}
        trainer_params["trainer"] = trainer.as_dict()

        # extra params
        trainer_params["trainer"]["share_dataset"] = use_shared_dataset

        # TODO: we set a random seed for each trainer
        #       as a committee will be trained
        #       it changes the trainer's random state as well...
        trainer_random_seed = np.random.randint(0, 10000)
        trainer_params["trainer"]["random_seed"] = trainer_random_seed
        trainer.set_rng(seed=trainer_random_seed)

        trainer_params["init_model"] = init_model
        trainer_params["dataset"] = dataset.as_dict()

        return trainer_params

    def run(self, dataset, size: int = 1, init_models=None, *args, **kwargs) -> None:
        """"""
        super().run(*args, **kwargs)
        if init_models is None:
            init_models = [None for i in range(size)]
        assert (
            len(init_models) == size
        ), "The number of init models is inconsistent with size."

        trainer = self.trainer
        scheduler = self.scheduler

        self._print(f"{trainer.random_seed = }")

        # - dump some metadata in case of resubmit
        meta_path = self._directory / "_meta"
        meta_path.mkdir(parents=True, exist_ok=True)

        meta_params = {}
        meta_params["dataset"] = dataset.as_dict()
        meta_params["size"] = size
        if init_models is not None:  # assume it is a List of path
            meta_params["init_models"] = [str(x) for x in init_models]
        else:
            meta_params["init_models"] = init_models
        with open(meta_path / "info.yaml", "w") as fopen:
            yaml.safe_dump(meta_params, fopen)

        # - check whether share a dataset?
        if size > 1 and self._share_dataset:
            dataset_path = self.directory / "shared_dataset"
            if not dataset_path.exists():
                self._print("prepare a shared dataset...")
                if hasattr(trainer, "_prepare_dataset"):
                    trainer.directory = dataset_path  # NOTE: only for creating dataset
                    dataset = trainer._prepare_dataset(dataset, *args, **kwargs)
                    self._print(f"{dataset =}")
                    with open(dataset_path/"dataset.yaml", "w") as fopen:
                        yaml.safe_dump(dataset.as_dict(), fopen)
                else:
                    self._print(
                        f"{trainer.__class__.__name__} does not support a shared dataset."
                    )
            else:
                # NOTE: sometimes local trainer does not finish,
                #       we need to load the shared dataset
                self._print("shared dataset exists...")
                with open(dataset_path/"dataset.yaml", "r") as fopen:
                    dataset_params = yaml.safe_load(fopen)
                dataset = DatasetVariable(**dataset_params).value
                self._print(f"{dataset =}")
        else:
            self._print("trainers prepare their own datasets...")

        # - read metadata from file or database
        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            queued_jobs = database.search(Query().queued.exists())
        queued_names = [q["gdir"][self.UUIDLEN + 1 :] for q in queued_jobs]

        # - run training
        for i in range(size):
            uid = str(uuid.uuid1())
            batch_name = f"{self.TRAIN_PREFIX}{i}"
            job_name = uid + "-" + batch_name
            wdir = self.directory / f"{self.TRAIN_PREFIX}{i}"
            if batch_name in queued_names:
                self._print(f"{job_name} at {self.directory.name} was submitted.")
                continue
            wdir.mkdir(parents=True, exist_ok=True)

            # - save trainer file for later reference
            # NOTE: YAML accepts only string path
            curr_init_model = init_models[i]
            if curr_init_model is not None:
                curr_init_model = str(curr_init_model)

            trainer_params = self._get_train_params(
                trainer, dataset, curr_init_model, self._share_dataset
            )
            with open(wdir / "trainer.yaml", "w") as fopen:
                yaml.dump(trainer_params, fopen)

            if self.scheduler.name == "local":
                # TODO: async ?
                self._print(f"training model at {wdir.name}...")
                trainer.directory = wdir
                trainer.train(dataset, init_model=curr_init_model)
            else:
                scheduler.job_name = job_name
                scheduler.script = wdir / "train.script"
                scheduler.user_commands = "gdp train {}\n".format(
                    str((wdir / "trainer.yaml").resolve())
                )
                scheduler.write()
                if self._submit:
                    self._print(f"{wdir.name}: {scheduler.submit()}")
                else:
                    self._print(f"{wdir.name} waits to submit.")

            # - update database
            with TinyDB(
                self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
            ) as database:
                _ = database.insert(
                    dict(
                        uid=uid,
                        gdir=job_name,
                        group_number=i,
                        wdir_names=[wdir.name],
                        queued=True,
                    )
                )

        return

    def inspect(self, resubmit=False, *args, **kwargs):
        """"""
        self._initialise(*args, **kwargs)
        self._debug(f"@@@{self.__class__.__name__}+inspect")

        running_jobs = self._get_running_jobs()

        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in running_jobs:
                doc_data = database.get(Query().gdir == job_name)
                uid = doc_data["uid"]
                wdir_names = doc_data["wdir_names"]

                self.scheduler.job_name = job_name
                self.scheduler.script = self.directory / "train.script"

                if self.scheduler.is_finished():  # NOTE: scheduler only checks job_name
                    # -- check if the job finished properly
                    is_finished = False
                    for x in wdir_names:
                        wdir_path = self.directory / x
                        if not wdir_path.exists():
                            break
                        else:
                            self.trainer.directory = wdir_path
                            if not self.trainer.read_convergence():
                                break
                    else:
                        is_finished = True
                    if is_finished:
                        database.update({"finished": True}, doc_ids=[doc_data.doc_id])
                    else:
                        if resubmit:
                            if self.scheduler.name != "local":
                                self._print(f"RESUBMIT: {str(self.trainer.directory)}")
                                if self._submit:
                                    self.scheduler.script = (
                                        self.trainer.directory / "train.script"
                                    )
                                    jobid = self.scheduler.submit()
                                    self._print(
                                        f"{job_name} is re-submitted with JOBID {jobid}."
                                    )
                                else:
                                    self._print(f"{job_name} waits to submit.")
                            else:
                                # NOTE: If the training runs on local,
                                #       database stores this job only when the training
                                #       is finished.
                                #       The codes below will only be performed when
                                #       self.inspect() is called without self.run() before.
                                # NOTE: Local job will automatically re-run when
                                #       self.run() is called.
                                self._print(f"RESUBMIT: {self.trainer.directory = }")
                                # self.trainer.train(dataset, init_model=init_models[i])
                        else:
                            warnings.warn(
                                "Trainer does not support re-submit.", UserWarning
                            )
                else:
                    self._print(f"{job_name} is running...")

        return

    def retrieve(self, include_retrieved: bool = False, *args, **kwargs):
        """Retrieve training results."""
        self.inspect(*args, **kwargs)
        self._debug(f"@@@{self.__class__.__name__}+retrieve")

        unretrieved_wdirs_ = []
        if not include_retrieved:
            unretrieved_jobs = self._get_unretrieved_jobs()
        else:
            unretrieved_jobs = self._get_finished_jobs()

        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                unretrieved_wdirs_.extend(
                    (self.directory / w).resolve() for w in doc_data["wdir_names"]
                )
        unretrieved_wdirs = unretrieved_wdirs_

        results = []
        if unretrieved_wdirs:
            unretrieved_wdirs = [pathlib.Path(x) for x in unretrieved_wdirs]
            # print("unretrieved_wdirs: ", unretrieved_wdirs)
            for p in unretrieved_wdirs:
                self.trainer.directory = p
                # NOTE: Due to yaml.safe_dump, we require path should be str
                results.append(str(self.trainer.freeze()))

        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                database.update({"retrieved": True}, doc_ids=[doc_data.doc_id])

        return results

    def _read_results(self, gdirs, *args, **kwargs):
        """"""
        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import argparse

import numpy as np
import matplotlib.pyplot as plt
plt.style.use("presentation")

from scipy.interpolate import griddata, interp1d
from scipy.optimize import minimize


def interpolate_string(a, b, t, npts: int=5, concat: bool=False):
    """"""
    # - use transition state
    if concat:
        npts_1 = int(npts/2.)
        npts_2 = npts - npts_1
        pts = np.vstack(
            [
                np.hstack([np.linspace(a[0], t[0], npts_1), np.linspace(t[0], b[0], npts_2)]),
                np.hstack([np.linspace(a[1], t[1], npts_1), np.linspace(t[1], b[1], npts_2)])
            ]
        ).T
    else:
        pts = np.vstack(
            [np.linspace(a[0], b[0], npts), np.linspace(a[1], b[1], npts)]
        ).T

    return pts

def prepare_data(data):
	""""""
	nx = np.size(np.unique(data[:, 0]))
	ny = np.size(np.unique(data[:, 1]))

	# Some people give input where the first column varies fastest. 
	# That is Fortran ordering, and you are liable to get things confused 
	# if you don't take this into account.
	order = "C"
	if data[0, 0] != data[1, 0]:
		order = "F"
	x = data[:, 0].reshape(nx, ny, order=order)
	y = data[:, 1].reshape(nx, ny, order=order)
	z = data[:, 2].reshape(nx, ny, order=order)
	#print(f"x: {x}")
	#print(f"y: {y}")

	# - basic input sanity check
	xdiff = np.diff(x, axis=0)
	ydiff = np.diff(y, axis=1)
	if (not np.all(np.abs(xdiff - xdiff[0]) < 1e-8)) or (not np.all(np.abs(ydiff - ydiff[0]) < 1e-8)):
		print("WARNING! The input data is not coming from an equally spaced grid. imshow will be subtly wrong as a result, as each pixel is assumed to represent the same area.")

	return x, y, z

def optimise_string(pts, gridpoints, gradx, grady, steps=100, stepsize=0.01):
	""""""
	npts = pts.shape[0]
	for i in range(steps):
		# - find gradient interpolation
		Dx = griddata(gridpoints, gradx.flatten(), (pts[:, 0], pts[:, 1]), method="linear")
		Dy = griddata(gridpoints, grady.flatten(), (pts[:, 0], pts[:, 1]), method="linear")
		h = np.amax(np.sqrt(np.square(Dx) + np.square(Dy)))

		# - evolve
		pts -= stepsize * np.vstack([Dx, Dy]).T / h

		# - reparameterize
		arclength = np.hstack([0, np.cumsum(np.linalg.norm(pts[1:] - pts[:-1], axis=1))])
		arclength /= arclength[-1]
		pts = np.vstack(
			[
				interp1d(arclength, pts[:, 0])(np.linspace(0, 1, npts)), 
				interp1d(arclength, pts[:, 1])(np.linspace(0, 1, npts))
			]
		).T

		# - save history?
		if i % 10 == 0:
			#print(i, np.sum(griddata(gridpoints, z.flatten(), (pts[:, 0], pts[:, 1]), method="linear")))
			print(i, np.sum(griddata(gridpoints, z.flatten(), (pts[:, 0], pts[:, 1]), method="linear")))

	return pts

def plot_pathways():
	""""""

	return

# ---
parser = argparse.ArgumentParser()
parser.add_argument("-f", "--file", help="free energy data", default="fes.dat", type=str)
parser.add_argument("-n", "--number", help="number of points", default=11, type=int)
parser.add_argument("-s", "--steps", help="number of steps", default=100, type=int)
parser.add_argument("-c", "--concat", help="interpolate mode", action="store_true")
args = parser.parse_args()

# Use the guessed basins to make the original string, 
# a linear interpolation between points a, t, and b.
npts = args.number #Number of intermediate points on the string.
steps = args.steps
concat = args.concat
fes_dat = args.file

# - preprocess data (load and reshape)
data = np.loadtxt(fes_dat)
print(f"data: {data.shape}")
data[:, 2] /= 96.485

x, y, z = prepare_data(data) # C-order 2D data

gridpoints = np.vstack([x.flatten(), y.flatten()]).T
print(f"grids: {gridpoints.shape}")
gradx, grady = np.gradient(z, np.unique(data[:, 0]), np.unique(data[:, 1]))

# Guesses for where the basins go. Since the input data is malformed, 
# you need to also provide an intermediate point t.
start_points = np.loadtxt("basin.dat").reshape(-1, 3, 2)
print(f"start_points: ")
print(start_points)

for i, (a, b, t) in enumerate(start_points):
	print(f"***** pathway {i} *****")
	pts = interpolate_string(a, b, t, npts=npts, concat=concat)

	# Evolve points so that they respond to the gradients. 
	# This is the "zero temperature string method"
	pts = optimise_string(pts, gridpoints, gradx, grady, steps=steps)
	ene = griddata(gridpoints, z.flatten(), (pts[:, 0], pts[:, 1]), method="linear")[:, np.newaxis]
	np.savetxt(f"./pmf-{i}.dat", np.hstack((np.linspace(0, 1, npts)[:, np.newaxis], pts, ene)), fmt="%.2f")

# ---
fig, ax = plt.subplots(1, 1, figsize=(12, 8))

ax.set_xlabel("CV 1")
ax.set_ylabel("CV 2")
#ax.set_xlim(0., 0.33)
#ax.set_ylim(0., 0.50)
cntr = ax.contourf(
    x, y, z, levels=10,
    #x, np.where(y < 0., y+1.0, y), z, levels=10,
    #cmap="RdBu"
)
fig.colorbar(cntr, ax=ax, label="Free Energy [eV]")
cn = ax.contour(
    x, y, z, #cmap="RdBu"
    #x, np.where(y < 0., y, y+1.0), z, levels=10,
)
plt.clabel(cn, inline=True, fontsize=10)

for i in range(start_points.shape[0]):
	pts = np.loadtxt(f"./pmf-{i}.dat")[:, 1:3]
	ax.plot(
		pts[:, 0], pts[:,1], linestyle="--", label=f"{i}"
	)
if start_points.shape[0] > 1:
    ax.legend()

fig.savefig("demo.png", transparent=False)
#fig.savefig("demo.png", transparent=True)

# --- 
fig, ax = plt.subplots(1, 1, figsize=(12, 8))
for i in range(start_points.shape[0]):
	pts = np.loadtxt(f"./pmf-{i}.dat")[:, 1:3]
	ax.plot(
		np.linspace(0, 1, npts), 
		griddata(gridpoints, z.flatten(), (pts[:, 0], pts[:, 1]), method="linear"),
	    marker="o", label=f"{i}"
	)
ax.legend()

ax.set_ylabel("Free Energy [eV]")
ax.set_xlabel("Reaction Coordinate")
fig.savefig("pmf.png")


if __name__ == "__main__":
	...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from .pathway import AseStringReactor
from .cp2k import Cp2kStringReactor
from .vasp import VaspStringReactor
from .grid import ZeroStringReactor


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import warnings

import numpy as np 
import numpy.ma as ma

from ase import Atoms 
from ase.io import read, write 

from ase.neighborlist import NeighborList 
from ase.neighborlist import natural_cutoffs 

from ase.calculators.calculator import (
    Calculator, all_changes, PropertyNotImplementedError
)
import deepmd.DeepPot as DeepPot


class DP(Calculator):
    name = "DP"
    implemented_properties = ["energy", "forces", "virial", "stress"]

    def __init__(self, model, label="DP", type_dict=None, **kwargs):
        Calculator.__init__(self, label=label, **kwargs)
        if isinstance(model, str):
            model = [model] # compatiblity
        self.dp_models = [] 
        for m in model:
            self.dp_models.append(DeepPot(m))
        if type_dict:
            self.type_dict=type_dict
        else:
            self.type_dict = dict(zip(self.dp.get_type_map(), range(self.dp.get_ntypes())))
    
    def prepare_input(self, atoms=None):
        coord = atoms.get_positions().reshape([1, -1])
        if sum(atoms.get_pbc())>0:
           self.pbc = True
           cell = atoms.get_cell().reshape([1, -1])
        else:
           self.pbc = False
           cell = None
        symbols = atoms.get_chemical_symbols()
        atype = [self.type_dict[k] for k in symbols]

        self.dp_input = {'coords': coord, 'cells': cell, 'atom_types': atype}

        return 

    def icalculate(self, dp_model, properties=["energy", "forces", "virial"], system_changes=all_changes):
        """"""
        results = {}
        e, f, v = dp_model.eval(**self.dp_input)
        results['energy'] = e[0][0]
        results['forces'] = f[0]
        results['virial'] = v[0].reshape(3,3)

        # convert virial into stress for lattice relaxation
        if "stress" in properties:
            if self.pbc:
                # the usual convention (tensile stress is positive)
                # stress = -virial / volume
                stress = -0.5*(v[0].copy()+v[0].copy().T) / atoms.get_volume()
                # Voigt notation 
                results['stress'] = stress.flat[[0,4,8,5,2,1]] 
            else:
                raise PropertyNotImplementedError
        
        return results
    
    def calculate(self, atoms=None, properties=["energy", "forces", "virial"], system_changes=all_changes):
        """"""
        self.prepare_input(atoms)
        all_results = []
        for dp_model in self.dp_models:
            cur_results = self.icalculate(dp_model) # return current results
            all_results.append(cur_results)
        
        if len(self.dp_models) == 1:
            self.results = all_results[0]
        else:
            # average results
            results = {}
            energy_array = [r['energy'] for r in all_results]
            results['energy'] = np.mean(energy_array)
            forces_array = np.array([r['forces'] for r in all_results])
            results['forces'] = np.mean(forces_array, axis=0)
            if 'stress' in properties:
                if self.pbc:
                    stress_array = np.array([r['stress'] for r in all_results])
                    results['stress'] = np.mean(stress_array, axis=0)
                else:
                    raise PropertyNotImplementedError 
            # estimate standard variance
            results['energy_stdvar'] = np.sqrt(np.var(energy_array))
            results['forces_stdvar'] = np.sqrt(np.var(forces_array, axis=0))
            self.results = results

        return 


def use_neighbors():

    #covlent_cutoffs = natural_cutoffs(atoms) 
    #nl = NeighborList(covlent_cutoffs)
    #nl.update(atoms)
    #indices, offsets = nl.get_neighbors(0)

    pass 


def cut_local_environment(atoms, centre_index, cutoff_radius=8.0, box_length=20.): 
    """""" 
    # get necessities 
    cell = atoms.cell
    frac_poses = atoms.get_scaled_positions()
    #print(frac_poses)
    #np.savetxt('n_large.txt',frac_poses)
    numbers = list(atoms.get_atomic_numbers())

    centre_frac_pos = frac_poses[centre_index]
    print('centre position: ', centre_frac_pos)
    centre_cart_pos = np.dot(frac_poses[centre_index], cell)
    # form a 3x3 box 
    offsets = np.array(
        [
            [0, 0, 0], [1, 0, 0], [-1, 0, 0], 
            [0, 1, 0], [1, 1, 0], [-1, 1, 0], 
            [0, -1, 0], [1, -1, 0], [-1, -1, 0], 
            [0, 0, 1], [1, 0, 1], [-1, 0, 1], 
            [0, 1, 1], [1, 1, 1], [-1, 1, 1], 
            [0, -1, 1], [1, -1, 1], [-1, -1, 1], 
            [0, 0, -1], [1, 0, -1], [-1, 0, -1], 
            [0, 1, -1], [1, 1, -1], [-1, 1, -1], 
            [0, -1, -1], [1, -1, -1], [-1, -1, -1]
        ]
    )
    
    extended_frac_poses = np.concatenate(
        (offsets.reshape(27,1,3)+frac_poses.reshape(1,-1,3)), axis=0
    )
    #print(extended_frac_poses.shape)
    extended_numbers = np.array(numbers*27)
    #print(extended_numbers)
    
    # calculate interatomic distances 
    frac_vectors = extended_frac_poses - centre_frac_pos 
    distances = np.linalg.norm(np.dot(frac_vectors, cell), axis=1)
    
    masked_distances = ma.masked_greater_equal(distances, cutoff_radius)
    #print(masked_distances.mask)
    
    # form a cluster 
    dis_mask = masked_distances.mask
    valid_positions = np.dot(
        np.array([extended_frac_poses[i] for i, b in enumerate(dis_mask) if not b]), cell
    )
    valid_positions = valid_positions - centre_cart_pos + np.ones(3)*box_length/2. # centre the cluster 
    #print(valid_positions.shape)
    valid_numbers = ma.array(extended_numbers, mask=masked_distances.mask).compressed() 
    #print(valid_numbers.shape)
    if box_length < 2*(cutoff_radius+5.):
        warnings.warn('The box may be too small for the cluster.', UserWarning)
    cluster_atoms = Atoms(
        numbers=valid_numbers, positions=valid_positions, cell=np.eye(3)*box_length
    )

    return cluster_atoms

def find_maxfstd(forces_stdvar, ifrac=False):
    """"""
    if ifrac:
        max_idx = np.argmax(forces_stdvar)
        irow = max_idx // 3
        icol = max_idx % 3
        atom_idx = irow
        max_fstdvar = forces_stdvar[irow,icol]
    else:
        tf_stdvar = np.linalg.norm(np.power(forces_stdvar,2), axis=1)
        atom_idx = np.argmax(tf_stdvar)
        max_fstdvar = tf_stdvar[atom_idx]

    return atom_idx, max_fstdvar

if __name__ == '__main__': 
    # ===== parameters =====
    # read structures 
    frames = read('test.xyz', ':')

    # set calculator
    type_map = {'O': 0, 'Pt': 1}
    model = [
        '/users/40247882/projects/oxides/gdp-main/it-0003/ensemble/model-0/graph.pb', 
        '/users/40247882/projects/oxides/gdp-main/it-0003/ensemble/model-1/graph.pb', 
        '/users/40247882/projects/oxides/gdp-main/it-0003/ensemble/model-2/graph.pb', 
        '/users/40247882/projects/oxides/gdp-main/it-0003/ensemble/model-3/graph.pb'
    ]

    calc = DP(model=model, type_dict=type_map)

    # deviate and cleave 
    min_stdv, max_stdv = 0.00, 0.25
    cutoff_radius, box_length = 7.0, 20.

    # ===== end parameters =====

    # main loop
    for frame_idx, atoms in enumerate(frames):
        print('structure %4d' %frame_idx)
        calc.reset()
        atoms.calc = calc
        dummy = atoms.get_forces() # carry out one calculation
        # energy_stdvar = atoms.calc.results.get('energy_stdvar', None) # total energy stdvar
        forces_stdvar = atoms.calc.results.get('forces_stdvar', None) # shape (natoms,3)
        atom_idx, max_fstdvar = find_maxfstd(forces_stdvar, ifrac=True) # use frac or summation of xyz
        if min_stdv < max_fstdvar <  max_stdv:
            print(atom_idx, max_fstdvar, 'cut')
            cluster_atoms = cut_local_environment(atoms, atom_idx, cutoff_radius=cutoff_radius, box_length=box_length) 
            write('cut_stru-%s-%s.xyz' %(str(frame_idx).zfill(4), str(atom_idx).zfill(4)), cluster_atoms)
        else:
            print(atom_idx, max_fstdvar, 'skip')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
import uuid
import warnings
import yaml

from tinydb import Query, TinyDB

from .worker import AbstractWorker
from ..scheduler.scheduler import AbstractScheduler

"""Worker that manages expeditions.

Since an expedition is made up of several basic workers, this worker is a monitor that 
tracks its progress.

"""

class ExpeditionBasedWorker(AbstractWorker):

    #: Prefix of the expedition folder name.
    EXP_INDEX: str = "expedition"

    batchsize: int = 1

    _script_name: str = "run.script"

    def __init__(self, expedition, scheduler: AbstractScheduler, batchsize: int=1, directory=None) -> None:
        """"""
        super().__init__(directory)

        self.expedition = expedition
        self.scheduler = scheduler

        self.batchsize = batchsize
        self.wait_time = 60

        return

    def run(self, builder=None, *args, **kwargs) -> None:
        """"""
        super().run(*args, **kwargs)

        expedition = self.expedition
        scheduler = self.scheduler

        # - read metadata from file or database
        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            queued_jobs = database.search(Query().queued.exists())
        queued_names = [q["gdir"][self.UUIDLEN+1:] for q in queued_jobs]

        size = 1
        for i in range(size):
            uid = str(uuid.uuid1())
            batch_name = f"{self.EXP_INDEX}-{i}"
            job_name = uid + "-" + self.EXP_INDEX + "-" + f"{i}"
            wdir = self.directory / (self.EXP_INDEX + "-" + f"{i}")
            if batch_name in queued_names:
                self._print(f"{job_name} at {self.directory.name} was submitted.")
                continue
            wdir.mkdir(parents=True, exist_ok=True)

            if self.scheduler.name == "local":
                expedition.directory = wdir
                expedition.run()
            else:
                inp_fpath = (wdir/f"exp-{uid}.yaml").absolute()
                exp_params = expedition.as_dict()
                with open(inp_fpath, "w") as fopen:
                    yaml.safe_dump(exp_params, fopen)

                scheduler.job_name = job_name
                scheduler.script = wdir/f"{self._script_name}-{uid}"
                scheduler.user_commands = "gdp explore {} --wait {}".format(
                    str(inp_fpath), self.wait_time
                )
                scheduler.write()
                if self._submit:
                    self._print(f"{wdir.name}: {scheduler.submit()}")
                else:
                    self._print(f"{wdir.name} waits to submit.")

            # - update database
            with TinyDB(
                self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
            ) as database:
                _ = database.insert(
                    dict(
                        uid = uid,
                        gdir=job_name, 
                        group_number=i, 
                        wdir_names=[wdir.name], 
                        queued=True
                    )
                )

        return
    
    def inspect(self, resubmit=False, *args, **kwargs):
        """"""
        self._initialise(*args, **kwargs)
        self._debug(f"~~~{self.__class__.__name__}+inspect")

        running_jobs = self._get_running_jobs()

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in running_jobs:
                doc_data = database.get(Query().gdir == job_name)
                uid = doc_data["uid"]
                wdir_names = doc_data["wdir_names"]

                self.scheduler.job_name = job_name
                self.scheduler.script = self.directory/f"{self._script_name}-{uid}"

                if self.scheduler.is_finished():
                    # -- check if the job finished properly
                    is_finished = False
                    for x in wdir_names:
                        wdir_path = self.directory/x
                        if not wdir_path.exists():
                            break
                        else:
                            self.expedition.directory = wdir_path
                            if not self.expedition.read_convergence():
                                break
                    else:
                        is_finished = True
                    if is_finished:
                        database.update({"finished": True}, doc_ids=[doc_data.doc_id])
                        #self._print(f"{job_name} finished.")
                    else:
                        warnings.warn("Exploration does not support re-submit.", UserWarning)
                        #if resubmit:
                        #    if self.scheduler.name != "local":
                        #        jobid = self.scheduler.submit()
                        #        self._print(f"{job_name} is re-submitted with JOBID {jobid}.")
                        #    else:
                        #        self._print(f"{job_name} tries to re-run.")
                        #        warnings.warn("Local scheduelr does not support re-run.", UserWarning)
                else:
                    self._print(f"{job_name} is running...")

        return
    
    def retrieve(self, include_retrieved: bool=False, *args, **kwargs):
        """"""
        #raise NotImplementedError(f"{self.__class__.__name__}")
        self.inspect(*args, **kwargs)
        self._debug(f"~~~{self.__class__.__name__}+retrieve")

        unretrieved_wdirs_ = []
        if not include_retrieved:
            unretrieved_jobs = self._get_unretrieved_jobs()
        else:
            unretrieved_jobs = self._get_finished_jobs()

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                unretrieved_wdirs_.extend(
                    (self.directory/w).resolve() for w in doc_data["wdir_names"]
                )
            unretrieved_wdirs = unretrieved_wdirs_

        workers = []
        if unretrieved_wdirs:
            unretrieved_wdirs = [pathlib.Path(x) for x in unretrieved_wdirs]
            self._debug(f"unretrieved_wdirs: {unretrieved_wdirs}")
            for p in unretrieved_wdirs:
                self.expedition.directory = p
                workers.extend(self.expedition.get_workers())

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                database.update({"retrieved": True}, doc_ids=[doc_data.doc_id])
        
        return workers


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import itertools

import numpy as np

from ..data.array import AtomsNDArray
from .selector import AbstractSelector


def convert_string_to_indices(indstr: str, length: int, convention="py"):
    """"""
    # NOTE: If the input is a valid number, we need convert it to str first.
    indstr = str(indstr)

    def _convert_string(x: str):
        """"""
        try:
            x = int(x)
        except:  # == ""
            x = None

        return x

    # print(f"indstr: {indstr}")
    indices = list(range(length))
    selected_indices = []
    for x in indstr.strip().split():
        curr_range = x.split(":")
        if len(curr_range) == 1:
            x = _convert_string(curr_range[0])
            if x is None:  # == ""
                start, stop, step = None, None, None
            else:
                if x >= 0:
                    start, stop, step = x, x + 1, None
                else:
                    start, stop, step = x, x - 1, -1
        elif len(curr_range) == 2:
            x, y = _convert_string(curr_range[0]), _convert_string(curr_range[1])
            start, stop, step = x, y, None
        elif len(curr_range) == 3:
            x, y, z = (
                _convert_string(curr_range[0]),
                _convert_string(curr_range[1]),
                _convert_string(curr_range[2]),
            )
            start, stop, step = x, y, z
        else:
            raise RuntimeError(f"Fail to parse the index string {x}.")
        # print(f"slice: {slice(start, stop, step)}")
        selected_indices.extend(indices[slice(start, stop, step)])

    return selected_indices


class LocateSelector(AbstractSelector):

    name = "locate"

    default_parameters = dict(
        indices=":"  # can be single integer, string or a List of integers
    )

    def __init__(self, directory="./", axis=None, *args, **kwargs) -> None:
        super().__init__(directory, axis, *args, **kwargs)

        return

    def _mark_structures(self, data: AtomsNDArray, *args, **kwargs) -> None:
        """"""
        super()._mark_structures(data, *args, **kwargs)

        axis = self.axis
        if axis < 0:
            axis = data.ndim + axis
        indices = self.indices

        # This is similar to np.take_along_axis
        if data.ndim > 0:
            marker_groups = {}
            for k, v in itertools.groupby(
                data.markers,
                key=lambda x: [x[i] for i in range(data.ndim) if i != axis],
            ):
                k = tuple(k)
                if k in marker_groups:
                    marker_groups[k].extend(list(v))
                else:
                    marker_groups[k] = list(v)
            selected_markers = []
            for k, v in marker_groups.items():
                v = sorted(np.array(v).tolist())
                selected_markers.extend(
                    [v[i] for i in convert_string_to_indices(self.indices, len(v))]
                )
            # self._print(f"selected_markers: {selected_markers}")
        else:
            raise RuntimeError(
                f"Locator does not support array dimension with {data.ndim}"
            )

        data.markers = selected_markers

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import time
import pickle
from typing import List
from pathlib import Path

import networkx as nx

from joblib import Parallel, delayed

"""Some methods to compare graph.
"""

#: Handles isomorphism for bonds.
bond_match = nx.algorithms.isomorphism.categorical_edge_match("bond", "")

#: Handles isomorphism for atoms with regards to perodic boundary conditions.
ads_match = nx.algorithms.isomorphism.categorical_node_match(["index", "ads"], [-1, False]) 

# - serial algorithms
def get_unique_environments_based_on_bonds(chem_envs: List[nx.Graph]) -> List[int]:
    """Find unique environments based on graph edges.

    This method compares one graph with another.

    Args:
        chem_envs: List of graph representation of an atom.

    Returns:
        Indices of unique graphs.
    
    """

    nsites = len(chem_envs)

    unique_indices = [0]
    for i in range(1,nsites):
        #print("u: ", unique_indices)
        for j in unique_indices:
            env_i, env_j = chem_envs[i], chem_envs[j]
            #print(i, env_i, j, env_j)
            #if nx.algorithms.isomorphism.is_isomorphic(env_i, env_j, edge_match=bond_match, node_match=ads_match):
            if nx.algorithms.isomorphism.is_isomorphic(env_i, env_j, edge_match=bond_match):
                #print(f"{i} == {j}")
                break
        else:
            unique_indices.append(i)

    return unique_indices

def get_unique_environments_based_on_nodes_and_edges(chem_envs):
    """Unique adsorbates. Get unique adsorbate chemical environment.
    
    Removes duplicate adsorbates which occur when PBCs detect the same 
    adsorbate in two places. Each adsorbate graph has its atomic index checked 
    to detect when PBC has created duplicates.

    Args:
        chem_env_groups (list[networkx.Graph]): Adsorbate/environment graphs

    Returns:
        list[networkx.Graph]: The unique adsorbate graphs

    """
    # Keep track of known unique adsorbate graphs
    unique = []
    for env in chem_envs:
        for unique_env in unique:
            if nx.algorithms.isomorphism.is_isomorphic(
                env, unique_env, edge_match=bond_match, node_match=ads_match
            ):
                break
        else: # Was unique
            unique.append(env)

    return unique

def compare_chem_envs(chem_envs1, chem_envs2):
    """Compares two sets of chemical environments to see if they are the same
    in chemical identity.  Useful for detecting if two sets of adsorbates are
    in the same configurations.

    Args:
        chem_envs1 (list[networkx.Graph]): A list of chemical environments, 
                                           including duplicate adsorbates
        chem_envs2 (list[networkx.Graph]): A list of chemical environments, 
                                           including duplicate adsorbates

    Returns:
        bool: Is there a matching graph (site / adsorbate) for each graph?

    """
    # Do they have the same number of adsorbates?
    if len(chem_envs1) != len(chem_envs2):
        return False

    envs_copy = chem_envs2[:] # Make copy of list

    # Check if chem_envs1 matches chem_envs2 by removing from envs_copy
    for env1 in chem_envs1: 
        for env2 in envs_copy:
            if nx.algorithms.isomorphism.is_isomorphic(env1, env2, edge_match=bond_match):
                # Remove this from envs_copy and move onto next env in chem_envs1
                envs_copy.remove(env2)
                break

    # Everything should have been removed from envs_copy if everything had a match
    if len(envs_copy) > 0:
        return False

    return True

def unique_chem_envs(chem_envs_groups, metadata=None, verbose=False):
    """Given a list of chemical environments, find the unique
    environments and keep track of metadata if required.

    This function exists largely to help with unique site detection
    but its performance will scale badly with extremely large numbers
    of chemical environments to check.  This can be split into parallel
    jobs.

    Args:
        chem_env_groups (list[list[networkx.Graph]]): 
            Chemical environments to compare against each other
        metadata (list[object]): 
            Corresponding metadata to keep with each chemical environment

    Returns:
        list[list[list[networkx.Graph]]]: A list of unique chemical environments 
                                          with their duplicates
        list[list[object]]: A matching list of metadata
    """
    # print("*** org algo for chem_env cmp***")
    # Error checking, this should never really happen
    #print("chem_envs: ", chem_envs_groups)
    if len(chem_envs_groups) == 0:
        return [[],[]]

    # We have metadata to manage
    if metadata is not None:
        if len(chem_envs_groups) != len(metadata):
            raise ValueError("Metadata must be the same length as\
                              the number of chem_envs_groups")
    
    # No metadata to keep track of
    if metadata is None:
        metadata = [None] * len(chem_envs_groups)

    # Keep track of known unique environments
    unique = []

    for index, env in enumerate(chem_envs_groups):
        #st = time.time()
        for index2, (unique_indices, unique_env) in enumerate(unique):
            if verbose:
                print("Checking for uniqueness {:05d}/{:05d} {:05d}/{:05d}".format(index+1, len(chem_envs_groups), index2, len(unique)), end='\r')
            if compare_chem_envs(env, unique_env):
                unique_indices.append(index)
                break
        else: # Was unique
            if verbose:
                print("")
            unique.append(([index], env))
        #et = time.time()
        #print("used time ", et-st, " for ", index, " for ", len(unique))
    
    # Zip trick to split into two lists to return
    # unique_envs, unique_groups
    # unique_groups = [[index,metadata],[]]
    return zip(*[(env, [metadata[index] for index in indices]) for (indices, env) in unique])

# - an algorithm for parallel comparison of graph groups
def new_unique_chem_envs(extra_info, chem_envs_groups, env_indices, verbose=False):
    """"""
    if verbose:
        print("*** new org algo for chem_env cmp***")
    # Error checking, this should never really happen
    #print("chem_envs: ", chem_envs_groups)
    if len(chem_envs_groups) == 0:
        return [[],[]]

    # Keep track of known unique environments
    unique = []

    for index in env_indices:
        st = time.time()
        env = chem_envs_groups[index]

        # NOTE: use a flag since there may continue all and not break
        found_similar = False
        for ug_idx, (unique_indices, ur_idx) in enumerate(unique):
            # check if already compared
            if ur_idx in extra_info[index]["others"]:
                continue
            # add reclusive
            extra_info[index]["others"].append(ur_idx)
            extra_info[ur_idx]["others"].append(index)
            # actual compare
            if compare_chem_envs(env, chem_envs_groups[ur_idx]):
                found_similar = True
                unique_indices.append(index)
                extra_info[ur_idx]["similar"].append(index)
                #print(ur_idx, extra_info[ur_idx])
                if len(extra_info[index]["similar"]) > 0:
                    extra_info[ur_idx]["similar"].extend(extra_info[index]["similar"])
                    extra_info[index]["similar"] = []
                break

        if not found_similar: # Was unique
            unique.append(([index], index))

        et = time.time()
        if verbose:
            print("used time ", et-st, " for ", index, " for ", len(unique))
    
    return unique

def get_indices(env_indices, n_jobs):
    nframes = len(env_indices)
    size = nframes // n_jobs

    st_indices = [x*size for x in range(n_jobs)]
    end_indices = [x+size for x in st_indices]
    end_indices[-1] = nframes

    env_indices_splits = []
    for s, e in zip(st_indices, end_indices):
        env_indices_splits.append([env_indices[x] for x in range(s,e)])

    return env_indices_splits

def get_groups(extra_info, env_indices, nsplits):
    """"""
    nframes = len(env_indices)
    size = nframes // nsplits
    st_indices = [x*size for x in range(nsplits)]
    end_indices = [x+size for x in st_indices]
    end_indices[-1] = nframes

    temp_env_indices = env_indices.copy()
    env_indices_splits = []

    nvalid = 0
    while nvalid != nsplits-1:
        if len(temp_env_indices) < 1:
            break
        s, e = st_indices[nvalid], end_indices[nvalid]
        size = e - s
        #print("size: ", size)
        #print(len(temp_env_indices), temp_env_indices)
        first = temp_env_indices[0]
        del temp_env_indices[0]
        nrest = len(temp_env_indices)
        #mates = [x for x in temp_env_indices if x not in extra_info[first]["others"]]
        mates = []
        count = 0
        while len(mates) < size-1:
            if count >= nrest:
                break
            if temp_env_indices[count] not in extra_info[first]["others"]:
                mates.append(temp_env_indices[count])
            count += 1
        if len(mates) < 1:
            continue
        env_indices_splits.append([first]+mates)
        temp_env_indices = list(set(temp_env_indices).difference(mates))
        nvalid += 1

    # last group
    if len(temp_env_indices) > 0:
        env_indices_splits.append(temp_env_indices)

    return env_indices_splits

def merge_results(extra_info, chem_envs_groups, metadata):
    duplicates = []
    for i, info in enumerate(extra_info):
        if len(info["similar"]) > 0:
            duplicates.extend(info["similar"])

    unique_indices = []
    unique_envs, unique_groups = [], []
    for i, info in enumerate(extra_info):
        if i in duplicates:
            continue
        unique_indices.append(i)
        unique_envs.append(chem_envs_groups[i])
        u_indices = [i]
        if len(info["similar"]) > 0:
            u_indices += info["similar"]
        u_indices = sorted(u_indices)
        unique_groups.append([metadata[u] for u in u_indices])
    # TODO: resort

    return unique_indices, unique_envs, unique_groups


def paragroup_unique_chem_envs(chem_envs_groups, metadata=None, directory=Path.cwd(), n_jobs=1):
    """"""
    # - switch to plain version if only one job is used
    if n_jobs == 1:
        unique_envs, unique_groups = unique_chem_envs(chem_envs_groups, metadata)
        return unique_envs, unique_groups

    # - run parallel version, not always correct but can reduce most duplicates
    nframes = len(chem_envs_groups)

    directory = Path(directory)
    saved_name = "extra_info.pkl"
    saved_info = directory / "extra_info.pkl"
    if saved_info.exists():
        with open(saved_info, "rb") as fopen:
            extra_info = pickle.load(fopen)
        env_indices, unique_envs, unique_groups = merge_results(extra_info, chem_envs_groups, metadata)
        print("load cached comparasion info...")
    else:
        extra_info = [
            {"similar": [], "others": []} for i in range(nframes)
        ]
        env_indices = list(range(nframes))

    # determine atomic task size
    # NOTE: MAYBE A BUG? 400/64 may miss some comparasions
    # that is for too small atomic task
    #nsplits = 64
    nsplits = n_jobs*(nframes // 16 // n_jobs)
    if nsplits == 0:
        nsplits += 1
    print("nsplits: ", nsplits)
    
    with Parallel(
        n_jobs=n_jobs, 
        prefer="threads",
        #require="sharedmem"
    ) as para:
        st = time.time()
        # split data

        for icount in range(500):
            print(f"--- {icount} ---")
            print("input nuniques: ", len(env_indices))
            #print("nsplits: ", nsplits)
            #env_indices_splits = get_indices(env_indices, nsplits)
            env_indices_splits = get_groups(extra_info, env_indices, nsplits)
            print(env_indices_splits)

            # run para cmp
            ret = para(delayed(new_unique_chem_envs)(extra_info, chem_envs_groups, env_indices) for env_indices in env_indices_splits)

            et = time.time()
            print("split ", nsplits, " for count ", icount, " time: ", et - st)

            # run final cmp
            env_indices = []
            for x in ret:
                env_indices.extend([a[1] for a in x])
            env_indices = sorted(env_indices)
            print("output nuniques: ", len(env_indices))

            #print("--- extra info ---")
            #for i, x in enumerate(extra_info):
            #    if i in env_indices:
            #        print(i, sorted(x["others"]))
            
            # check diff
            temp_env_indices = env_indices.copy()
            for e_idx in temp_env_indices:
                compared_others = sorted([e_idx] + extra_info[e_idx]["others"])
                if set(compared_others).intersection(set(temp_env_indices)) != set(temp_env_indices):
                    break
            else:
                print("finished...")
                break
                
            if (icount+1) % 50 == 0:
                print("save temp comparasion info...")
                with open(directory/(saved_name+"-"+str(icount+1)), "wb") as fopen:
                    pickle.dump(extra_info, fopen)
            
        #unique = new_unique_chem_envs(extra_info, chem_envs_groups, env_indices)
        #et = time.time()
        #print("complete time: ", et - st)
    
    with open(saved_info, "wb") as fopen:
        pickle.dump(extra_info, fopen)

    st = time.time()

    unique_indices, unique_envs, unique_groups = merge_results(extra_info, chem_envs_groups, metadata)
        
    et = time.time()
    print("merge data time: ", et - st)

    return unique_envs, unique_groups

if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import List, Callable

from ase import Atoms
from ase.calculators.singlepoint import SinglePointCalculator


def copy_minimal_frames(prev_frames: List[Atoms]):
    """Copy atoms without extra information.

    Do not copy atoms.info since it is a dict and does not maitain order.

    """
    curr_frames, curr_info = [], []
    for prev_atoms in prev_frames:
        # - copy geometry
        curr_atoms = Atoms(
            symbols=copy.deepcopy(prev_atoms.get_chemical_symbols()),
            positions=copy.deepcopy(prev_atoms.get_positions()),
            cell=copy.deepcopy(prev_atoms.get_cell(complete=True)),
            pbc=copy.deepcopy(prev_atoms.get_pbc()),
        )
        curr_frames.append(curr_atoms)
        # - save info
        confid = prev_atoms.info.get("confid", -1)
        dynstep = prev_atoms.info.get("step", -1)
        prev_wdir = prev_atoms.info.get("wdir", "null")
        curr_info.append((confid, dynstep, prev_wdir))

    return curr_frames, curr_info


def make_clean_atoms(atoms_: Atoms, results: dict = None):
    """Create a clean atoms from the input."""
    atoms = Atoms(
        symbols=atoms_.get_chemical_symbols(),
        positions=atoms_.get_positions().copy(),
        cell=atoms_.get_cell().copy(),
        pbc=copy.deepcopy(atoms_.get_pbc()),
    )
    if results is not None:
        spc = SinglePointCalculator(atoms, **results)
        atoms.calc = spc

    return atoms


def parse_type_list(atoms):
    """parse type list for read and write structure of lammps"""
    # elements
    type_list = list(set(atoms.get_chemical_symbols()))
    type_list.sort()  # by alphabet

    return type_list


def get_composition_from_atoms(atoms):
    """"""
    from collections import Counter

    chemical_symbols = atoms.get_chemical_symbols()
    composition = Counter(chemical_symbols)
    sorted_composition = sorted(composition.items(), key=lambda x: x[0])

    return sorted_composition


def get_formula_from_atoms(atoms):
    """"""
    from collections import Counter

    chemical_symbols = atoms.get_chemical_symbols()
    composition = Counter(chemical_symbols)
    sorted_composition = sorted(composition.items(), key=lambda x: x[0])

    return "".join([str(k) + str(v) for k, v in sorted_composition])


if __name__ == "__main__":
    ...
